from __future__ import annotations

from ctypes import c_void_p
from typing import TYPE_CHECKING, Sequence

import pyglet
from pyglet.display.cocoa import CocoaCanvas
from pyglet.event import EventDispatcher
from pyglet.libs.darwin import AutoReleasePool, CGPoint, cocoapy
from pyglet.window import BaseWindow, DefaultMouseCursor, MouseCursor

from ...libs import darwin
from .pyglet_delegate import PygletDelegate
from .pyglet_textview import PygletTextView
from .pyglet_view import PygletView
from .pyglet_window import PygletToolWindow, PygletWindow
from .systemcursor import SystemCursor

if TYPE_CHECKING:
    from pyglet.gl.cocoa import CocoaContext

NSApplication = cocoapy.ObjCClass('NSApplication')
NSCursor = cocoapy.ObjCClass('NSCursor')
NSColor = cocoapy.ObjCClass('NSColor')
NSEvent = cocoapy.ObjCClass('NSEvent')
NSArray = cocoapy.ObjCClass('NSArray')
NSImage = cocoapy.ObjCClass('NSImage')
NSPasteboard = cocoapy.ObjCClass('NSPasteboard')

quartz = cocoapy.quartz
cf = cocoapy.cf


class CocoaMouseCursor(MouseCursor):
    gl_drawable = False

    def __init__(self, cursorName: str) -> None:
        # cursorName is a string identifying one of the named default NSCursors
        # e.g. 'pointingHandCursor', and can be sent as message to NSCursor class.
        self.cursorName = cursorName

    def set(self) -> None:
        cursor = getattr(NSCursor, self.cursorName)()
        cursor.set()


class CocoaWindow(BaseWindow):
    context: CocoaContext
    # NSWindow instance.
    _nswindow: darwin.ObjCInstance | None = None

    # Delegate object.
    _delegate: darwin.ObjCInstance | None = None

    # Window properties
    _mouse_platform_visible: bool = True
    _mouse_ignore_motion: bool = False

    # Flag set during close() method.
    _was_closed: bool = False

    # NSWindow style masks.
    _style_masks: dict[str, int] = {
        BaseWindow.WINDOW_STYLE_DEFAULT: cocoapy.NSTitledWindowMask |
                                         cocoapy.NSClosableWindowMask |
                                         cocoapy.NSMiniaturizableWindowMask,
        BaseWindow.WINDOW_STYLE_DIALOG: cocoapy.NSTitledWindowMask |
                                        cocoapy.NSClosableWindowMask,
        BaseWindow.WINDOW_STYLE_TOOL: cocoapy.NSTitledWindowMask |
                                      cocoapy.NSClosableWindowMask |
                                      cocoapy.NSUtilityWindowMask,
        BaseWindow.WINDOW_STYLE_BORDERLESS: cocoapy.NSBorderlessWindowMask,
        BaseWindow.WINDOW_STYLE_TRANSPARENT: cocoapy.NSTitledWindowMask |
                                         cocoapy.NSClosableWindowMask |
                                         cocoapy.NSMiniaturizableWindowMask,
        BaseWindow.WINDOW_STYLE_OVERLAY: cocoapy.NSBorderlessWindowMask,

    }

    def __init__(self, *args, **kwargs) -> None:  # noqa: ANN002, ANN003
        with AutoReleasePool():
            super().__init__(*args, **kwargs)

    def _recreate(self, changes: Sequence[str]) -> None:
        if 'context' in changes:
            self.context.set_current()

        if 'fullscreen' in changes and not self._fullscreen:  # leaving fullscreen
            self.screen.release_display()

        self._create()

    def _create(self) -> None:
        with AutoReleasePool():
            if self._nswindow:
                # The window is about the be recreated so destroy everything
                # associated with the old window, then destroy the window itself.
                nsview = self.canvas.nsview
                self.canvas = None
                self._nswindow.orderOut_(None)
                self._nswindow.close()
                self.context.detach()
                self._nswindow.release()
                self._nswindow = None
                nsview.release()
                self._delegate.release()
                self._delegate = None

            # Determine window parameters.
            if pyglet.options.dpi_scaling == "real":
                screen_scale = self.screen.get_scale()
                w, h = self.get_requested_size()
                width, height = w / screen_scale, h / screen_scale
            else:
                width, height = self._width, self._height

            content_rect = cocoapy.NSMakeRect(0, 0, width, height)
            WindowClass = PygletWindow
            if self._fullscreen:
                style_mask = cocoapy.NSBorderlessWindowMask
            else:
                if self._style not in self._style_masks:
                    self._style = self.WINDOW_STYLE_DEFAULT
                style_mask = self._style_masks[self._style]
                if self._resizable:
                    style_mask |= cocoapy.NSResizableWindowMask
                if self._style == BaseWindow.WINDOW_STYLE_TOOL:
                    WindowClass = PygletToolWindow

            # First create an instance of our NSWindow subclass.

            # FIX ME:
            # Need to use this initializer to have any hope of multi-monitor support.
            # But currently causes problems on Mac OS X Lion.  So for now, we initialize the
            # window without including screen information.
            #
            # self._nswindow = WindowClass.alloc().initWithContentRect_styleMask_backing_defer_screen_(
            #     content_rect,           # contentRect
            #     style_mask,             # styleMask
            #     NSBackingStoreBuffered, # backing
            #     False,                  # defer
            #     self.screen.get_nsscreen())  # screen

            self._nswindow = WindowClass.alloc().initWithContentRect_styleMask_backing_defer_(
                content_rect,  # contentRect
                style_mask,  # styleMask
                cocoapy.NSBackingStoreBuffered,  # backing
                False)  # defer

            if self._fullscreen:
                # BUG: I suspect that this doesn't do the right thing when using
                # multiple monitors (which would be to go fullscreen on the monitor
                # where the window is located).  However I've no way to test.
                blackColor = NSColor.blackColor()
                self._nswindow.setBackgroundColor_(blackColor)
                self._nswindow.setOpaque_(True)
                self.screen.capture_display()
                self._nswindow.setLevel_(quartz.CGShieldingWindowLevel())
                self.context.set_full_screen()
                self._center_window()
                self._mouse_in_window = True
            else:
                self._set_nice_window_location()
                self._mouse_in_window = self._mouse_in_content_rect()

            # Then create a view and set it as our NSWindow's content view.
            self._nsview = PygletView.alloc().initWithFrame_cocoaWindow_(content_rect, self)
            self._nsview.setWantsBestResolutionOpenGLSurface_(True)

        if not self._fullscreen:
            if self._style in ("transparent", "overlay"):
                self._nswindow.setOpaque_(False)
                self._nswindow.setBackgroundColor_(NSColor.clearColor())
                self._nswindow.setHasShadow_(False)

                if self._style == "overlay":
                    self.set_mouse_passthrough(True)
                    self._nswindow.setLevel_(cocoapy.NSStatusWindowLevel)

            self._nswindow.setContentView_(self._nsview)
            self._nswindow.makeFirstResponder_(self._nsview)

            # Create a canvas with the view as its drawable and attach context to it.
            self.canvas = CocoaCanvas(self.display, self.screen, self._nsview)
            self.context.attach(self.canvas)

            # Configure the window.
            self._nswindow.setAcceptsMouseMovedEvents_(True)

            # Required as it may cause a segfault after closing a Window, mostly due to NSTextView use.
            self._nswindow.setReleasedWhenClosed_(False)
            self._nswindow.useOptimizedDrawing_(True)
            self._nswindow.setPreservesContentDuringLiveResize_(False)

            # Set the delegate.
            self._delegate = PygletDelegate.alloc().initWithWindow_(self)

            # Configure CocoaWindow.
            self.set_caption(self._caption)
            if self._minimum_size is not None:
                self.set_minimum_size(*self._minimum_size)
            if self._maximum_size is not None:
                self.set_maximum_size(*self._maximum_size)

            self._dpi = self._get_dpi_desc()

            if self._file_drops:
                array = NSArray.arrayWithObject_(cocoapy.NSPasteboardTypeURL)
                self._nsview.registerForDraggedTypes_(array)

            self.context.update_geometry()
            self.switch_to()
            self.set_vsync(self._vsync)
            self.set_visible(self._visible)

    def _get_dpi_desc(self) -> int:
        if pyglet.options.dpi_scaling in ("scaled", "stretch", "platform") and self._nswindow:
            desc = self._nswindow.deviceDescription()
            rsize = desc.objectForKey_(darwin.NSDeviceResolution).sizeValue()
            return int(rsize.width)

        return 72

    @property
    def scale(self) -> float:
        """The scale of the window factoring in DPI.

        Read only.
        """
        if pyglet.options.dpi_scaling in ("scaled", "stretch", "platform") and self._nswindow:
            return self._nswindow.backingScaleFactor()

        return 1.0

    def _get_mouse_scale(self) -> float:
        """The mouse scale factoring in the DPI.

        On Mac, this is always 1.0.
        """
        return 1.0

    def set_mouse_passthrough(self, state: bool) -> None:
        with AutoReleasePool():
            self._nswindow.setIgnoresMouseEvents_(state)

    def _set_nice_window_location(self) -> None:
        # Construct a list of all visible windows that aren't us.
        visible_windows = [win for win in pyglet.app.windows if
                           win is not self and
                           win._nswindow and  # noqa: SLF001
                           win._nswindow.isVisible()]  # noqa: SLF001
        # If there aren't any visible windows, then center this window.
        if not visible_windows:
            self._center_window()
        # Otherwise, cascade from last window in list.
        else:
            point = visible_windows[-1]._nswindow.cascadeTopLeftFromPoint_(cocoapy.NSZeroPoint)  # noqa: SLF001
            self._nswindow.cascadeTopLeftFromPoint_(point)

    def _center_window(self) -> None:
        # [NSWindow center] does not move the window to a true center position
        # and also always moves the window to the main display.
        x = self.screen.x + int((self.screen.width - self._width) // 2)
        y = self.screen.y + int((self.screen.height - self._height) // 2)
        self._nswindow.setFrameOrigin_(cocoapy.NSPoint(x, y))

    def close(self) -> None:
        # If we've already gone through this once, don't do it again.
        if self._was_closed:
            return

        with AutoReleasePool():
            # Restore cursor visibility
            self.set_mouse_platform_visible(True)
            self.set_exclusive_mouse(False)
            self.set_exclusive_keyboard(False)

            # Remove window from display and remove its view.
            self._nswindow.orderOut_(None)

            # Restore screen mode. This also releases the display
            # if it was captured for fullscreen mode.
            self.screen.restore_mode()

            # Remove the delegate object
            if self._delegate:
                self._nswindow.setDelegate_(None)
                self._delegate.release()
                self._delegate = None

            # Remove view from canvas and then remove canvas.
            if self.canvas:
                self.canvas.nsview = None
                self.canvas = None

            if self._nsview:
                self._nswindow.setContentView_(None)
                self._nsview.release()
                self._nsview = None

            self._nswindow.close()
            self._nswindow = None

            # Dispatch any events that may be queued up, which includes deallocations.
            self._poll_app_events()

            # Do this last, so that we don't see white flash
            # when exiting application from fullscreen mode.
            super().close()

            self._was_closed = True

    def switch_to(self) -> None:
        if self.context:
            self.context.set_current()

    def flip(self) -> None:
        self.draw_mouse_cursor()
        if self.context:
            self.context.flip()

    def _poll_app_events(self):
        with AutoReleasePool():
            while True:
                NSApp = NSApplication.sharedApplication()

                event = NSApp.nextEventMatchingMask_untilDate_inMode_dequeue_(
                    cocoapy.NSAnyEventMask, None, cocoapy.NSDefaultRunLoopMode, True)

                if event is None:
                    break

                NSApp.sendEvent_(event)

    def dispatch_events(self) -> None:
        self._allow_dispatch_event = True
        # Process all pyglet events.
        self.dispatch_pending_events()
        event = True

        # Dequeue and process all of the pending Cocoa events.
        with AutoReleasePool():
            NSApp = NSApplication.sharedApplication()
            while event and self._nswindow and self._context:
                event = NSApp.nextEventMatchingMask_untilDate_inMode_dequeue_(
                    cocoapy.NSAnyEventMask, None, cocoapy.NSEventTrackingRunLoopMode, True)

                if event:
                    event_type = event.type()
                    # Pass on all events.
                    NSApp.sendEvent_(event)
                    # And resend key events to special handlers.
                    if event_type == cocoapy.NSKeyDown and not event.isARepeat():
                        NSApp.sendAction_to_from_(cocoapy.get_selector('pygletKeyDown:'), None, event)
                    elif event_type == cocoapy.NSKeyUp:
                        NSApp.sendAction_to_from_(cocoapy.get_selector('pygletKeyUp:'), None, event)
                    elif event_type == cocoapy.NSFlagsChanged:
                        NSApp.sendAction_to_from_(cocoapy.get_selector('pygletFlagsChanged:'), None, event)
                    NSApp.updateWindows()

        self._allow_dispatch_event = False

    def dispatch_pending_events(self) -> None:
        while self._event_queue:
            event = self._event_queue.pop(0)
            EventDispatcher.dispatch_event(self, *event)

    def set_caption(self, caption: str) -> None:
        self._caption = caption
        if self._nswindow is not None:
            self._nswindow.setTitle_(cocoapy.get_NSString(caption))

    def set_icon(self, *images: pyglet.image.ImageData) -> None:
        # Only use the biggest image from the list.
        max_image = images[0]
        for img in images:
            if img.width > max_image.width and img.height > max_image.height:
                max_image = img

        # Grab image data from pyglet image.
        image = max_image.get_image_data()
        fmt = 'ARGB'
        bytesPerRow = len(fmt) * image.width
        data = image.get_data(fmt, -bytesPerRow)

        # Use image data to create a data provider.
        # Using CGDataProviderCreateWithData crashes PyObjC 2.2b3, so we create
        # a CFDataRef object first and use it to create the data provider.
        cfdata = c_void_p(cf.CFDataCreate(None, data, len(data)))

        provider = c_void_p(quartz.CGDataProviderCreateWithCFData(cfdata))

        colorSpace = c_void_p(quartz.CGColorSpaceCreateDeviceRGB())

        # Then create a CGImage from the provider.
        cgimage = c_void_p(quartz.CGImageCreate(
            image.width, image.height, 8, 32, bytesPerRow,
            colorSpace,
            cocoapy.kCGImageAlphaFirst,
            provider,
            None,
            True,
            cocoapy.kCGRenderingIntentDefault))

        if not cgimage:
            return

        cf.CFRelease(cfdata)
        quartz.CGDataProviderRelease(provider)
        quartz.CGColorSpaceRelease(colorSpace)

        # Turn the CGImage into an NSImage.
        size = cocoapy.NSMakeSize(image.width, image.height)
        nsimage = NSImage.alloc().initWithCGImage_size_(cgimage, size)
        if not nsimage:
            return

        # And finally set the app icon.
        NSApp = NSApplication.sharedApplication()
        NSApp.setApplicationIconImage_(nsimage)
        nsimage.release()

    def get_location(self) -> tuple[int, int]:
        window_frame = self._nswindow.frame()
        rect = self._nswindow.contentRectForFrameRect_(window_frame)
        screen_frame = self._nswindow.screen().frame()
        screen_width = int(screen_frame.size.width)  # noqa: F841
        screen_height = int(screen_frame.size.height)
        return int(rect.origin.x), int(screen_height - rect.origin.y - rect.size.height)

    def set_location(self, x: int, y: int) -> None:
        window_frame = self._nswindow.frame()
        rect = self._nswindow.contentRectForFrameRect_(window_frame)
        screen_frame = self._nswindow.screen().frame()
        screen_width = int(screen_frame.size.width)  # noqa: F841
        screen_height = int(screen_frame.size.height)
        origin = cocoapy.NSPoint(x, screen_height - y - rect.size.height)
        self._nswindow.setFrameOrigin_(origin)

    def get_size(self) -> tuple[int, int]:
        if pyglet.options.dpi_scaling != "stretch":
            return self.get_framebuffer_size()

        return self._width, self._height

    def get_framebuffer_size(self) -> tuple[int, int]:
        view = self.context._nscontext.view()
        bounds = view.bounds()
        bounds = view.convertRectToBacking_(bounds)
        return int(bounds.size.width), int(bounds.size.height)

    def set_size(self, width: int, height: int) -> None:
        super().set_size(width, height)

        if pyglet.options.dpi_scaling == "real":
            screen_scale = self._nswindow.backingScaleFactor()
            frame_width, frame_height = width // screen_scale, height // screen_scale
        else:
            frame_width, frame_height = width, height

        self._set_frame_size(frame_width, frame_height)
        self.dispatch_event('_on_internal_resize', width, height)

    def _set_frame_size(self, width: int, height: int) -> None:
        # Move frame origin down so that top-left corner of window doesn't move.
        window_frame = self._nswindow.frame()
        rect = self._nswindow.contentRectForFrameRect_(window_frame)
        rect.origin.y += rect.size.height - height
        rect.size.width = width
        rect.size.height = height
        new_frame = self._nswindow.frameRectForContentRect_(rect)
        # The window background flashes when the frame size changes unless it's
        # animated, but we can set the window's animationResizeTime to zero.
        is_visible = self._nswindow.isVisible()
        self._nswindow.setFrame_display_animate_(new_frame, True, is_visible)

    def set_minimum_size(self, width: int, height: int) -> None:
        super().set_minimum_size(width, height)

        if self._nswindow is not None:
            ns_minimum_size = cocoapy.NSSize(*self._minimum_size)
            self._nswindow.setContentMinSize_(ns_minimum_size)

    def set_maximum_size(self, width: int, height: int) -> None:
        super().set_maximum_size(width, height)

        if self._nswindow is not None:
            ns_maximum_size = cocoapy.NSSize(*self._maximum_size)
            self._nswindow.setContentMaxSize_(ns_maximum_size)

    def activate(self) -> None:
        if self._nswindow is not None:
            NSApp = NSApplication.sharedApplication()
            NSApp.activateIgnoringOtherApps_(True)
            self._nswindow.makeKeyAndOrderFront_(None)

    def set_visible(self, visible: bool = True) -> None:
        super().set_visible(visible)

        if self._nswindow is not None:
            if visible:
                self.dispatch_event('_on_internal_resize', self._width, self._height)
                self.dispatch_event('on_show')
                self.dispatch_event('on_expose')
                self._nswindow.makeKeyAndOrderFront_(None)
            else:
                self._nswindow.orderOut_(None)

    def minimize(self) -> None:
        self._mouse_in_window = False
        if self._nswindow is not None:
            self._nswindow.miniaturize_(None)

    def maximize(self) -> None:
        if self._nswindow is not None:
            self._nswindow.zoom_(None)

    def set_vsync(self, vsync: bool) -> None:
        if pyglet.options['vsync'] is not None:
            vsync = pyglet.options['vsync']

        super().set_vsync(vsync)
        self.context.set_vsync(vsync)

    def _mouse_in_content_rect(self) -> bool:
        # Returns true if mouse is inside the window's content rectangle.
        # Better to use this method to check manually rather than relying
        # on instance variables that may not be set correctly.
        point = NSEvent.mouseLocation()
        window_frame = self._nswindow.frame()
        rect = self._nswindow.contentRectForFrameRect_(window_frame)
        return cocoapy.foundation.NSMouseInRect(point, rect, False)

    def set_mouse_platform_visible(self, platform_visible: int | None = None) -> None:
        # When the platform_visible argument is supplied with a boolean, then this
        # method simply sets whether or not the platform mouse cursor is visible.
        if platform_visible is not None:
            if platform_visible:
                SystemCursor.unhide()
            else:
                SystemCursor.hide()
        # But if it has been called without an argument, it turns into
        # a completely different function.  Now we are trying to figure out
        # whether or not the mouse *should* be visible, and if so, what it should
        # look like.
        else:
            # If we are in mouse exclusive mode, then hide the mouse cursor.
            if self._mouse_exclusive:
                SystemCursor.hide()
            # If we aren't inside the window, then always show the mouse
            # and make sure that it is the default cursor.
            elif not self._mouse_in_content_rect():
                NSCursor.arrowCursor().set()
                SystemCursor.unhide()
            # If we are in the window, then what we do depends on both
            # the current pyglet-set visibility setting for the mouse and
            # the type of the mouse cursor.  If the cursor has been hidden
            # in the window with set_mouse_visible() then don't show it.
            elif not self._mouse_visible:
                SystemCursor.hide()
            # If the mouse is set as a system-defined cursor, then we
            # need to set the cursor and show the mouse.
            # *** FIX ME ***
            elif isinstance(self._mouse_cursor, CocoaMouseCursor):
                self._mouse_cursor.set()
                SystemCursor.unhide()
            # If the mouse cursor is OpenGL drawable, then it we need to hide
            # the system mouse cursor, so that the cursor can draw itself.
            elif self._mouse_cursor.gl_drawable:
                SystemCursor.hide()
            # Otherwise, show the default cursor.
            else:
                NSCursor.arrowCursor().set()
                SystemCursor.unhide()

    def get_system_mouse_cursor(self, name: str) -> DefaultMouseCursor | CocoaMouseCursor:
        # It would make a lot more sense for most of this code to be
        # inside the CocoaMouseCursor class, but all of the CURSOR_xxx
        # constants are defined as properties of BaseWindow.
        if name == self.CURSOR_DEFAULT:
            return DefaultMouseCursor()
        cursors = {
            self.CURSOR_CROSSHAIR: 'crosshairCursor',
            self.CURSOR_HAND: 'pointingHandCursor',
            self.CURSOR_HELP: 'arrowCursor',
            self.CURSOR_NO: 'operationNotAllowedCursor',  # Mac OS 10.6
            self.CURSOR_SIZE: 'arrowCursor',
            self.CURSOR_SIZE_UP: 'resizeUpCursor',
            self.CURSOR_SIZE_UP_RIGHT: 'arrowCursor',
            self.CURSOR_SIZE_RIGHT: 'resizeRightCursor',
            self.CURSOR_SIZE_DOWN_RIGHT: 'arrowCursor',
            self.CURSOR_SIZE_DOWN: 'resizeDownCursor',
            self.CURSOR_SIZE_DOWN_LEFT: 'arrowCursor',
            self.CURSOR_SIZE_LEFT: 'resizeLeftCursor',
            self.CURSOR_SIZE_UP_LEFT: 'arrowCursor',
            self.CURSOR_SIZE_UP_DOWN: 'resizeUpDownCursor',
            self.CURSOR_SIZE_LEFT_RIGHT: 'resizeLeftRightCursor',
            self.CURSOR_TEXT: 'IBeamCursor',
            self.CURSOR_WAIT: 'arrowCursor',  # No wristwatch cursor in Cocoa
            self.CURSOR_WAIT_ARROW: 'arrowCursor',  # No wristwatch cursor in Cocoa
        }
        if name not in cursors:
            msg = f'Unknown cursor name "{name}"'
            raise RuntimeError(msg)
        return CocoaMouseCursor(cursors[name])

    def set_mouse_position(self, x: int, y: int, absolute: bool = False) -> None:
        if absolute:
            # If absolute, then x, y is given in global display coordinates
            # which sets (0,0) at top left corner of main display.  It is possible
            # to warp the mouse position to a point inside of another display.
            quartz.CGWarpMouseCursorPosition(CGPoint(x, y))
        else:
            # Window-relative coordinates: (x, y) are given in window coords
            # with (0,0) at bottom-left corner of window and y up.  We find
            # which display the window is in and then convert x, y into local
            # display coords where (0,0) is now top-left of display and y down.
            screenInfo = self._nswindow.screen().deviceDescription()
            displayID = screenInfo.objectForKey_(cocoapy.get_NSString('NSScreenNumber'))
            displayID = displayID.intValue()
            displayBounds = quartz.CGDisplayBounds(displayID)
            frame = self._nswindow.frame()
            windowOrigin = frame.origin
            x += windowOrigin.x
            y = displayBounds.size.height - windowOrigin.y - y
            quartz.CGDisplayMoveCursorToPoint(displayID, cocoapy.NSPoint(x, y))

    def set_exclusive_mouse(self, exclusive: bool = True) -> None:
        super().set_exclusive_mouse(exclusive)
        if exclusive:
            # Skip the next motion event, which would return a large delta.
            self._mouse_ignore_motion = True
            # Move mouse to center of window.
            frame = self._nswindow.frame()
            width, height = frame.size.width, frame.size.height
            self.set_mouse_position(width / 2, height / 2)
            quartz.CGAssociateMouseAndMouseCursorPosition(False)
        else:
            quartz.CGAssociateMouseAndMouseCursorPosition(True)

        # Update visibility of mouse cursor.
        self.set_mouse_platform_visible()

    def set_exclusive_keyboard(self, exclusive: bool = True) -> None:
        # http://developer.apple.com/mac/library/technotes/tn2002/tn2062.html
        # http://developer.apple.com/library/mac/#technotes/KioskMode/

        # BUG: System keys like F9 or command-tab are disabled, however
        # pyglet also does not receive key press events for them.

        # This flag is queried by window delegate to determine whether
        # the quit menu item is active.
        super().set_exclusive_keyboard(exclusive)

        if exclusive:
            # "Be nice! Don't disable force-quit!"
            #          -- Patrick Swayze, Road House (1989)
            options = cocoapy.NSApplicationPresentationHideDock | \
                      cocoapy.NSApplicationPresentationHideMenuBar | \
                      cocoapy.NSApplicationPresentationDisableProcessSwitching | \
                      cocoapy.NSApplicationPresentationDisableHideApplication
        else:
            options = cocoapy.NSApplicationPresentationDefault

        NSApp = NSApplication.sharedApplication()
        NSApp.setPresentationOptions_(options)

    def set_clipboard_text(self, text: str) -> None:
        with AutoReleasePool():
            pasteboard = NSPasteboard.generalPasteboard()

            pasteboard.clearContents()

            array = NSArray.arrayWithObject_(cocoapy.NSPasteboardTypeString)
            pasteboard.declareTypes_owner_(array, None)

            text_nsstring = cocoapy.get_NSString(text)

            pasteboard.setString_forType_(text_nsstring, cocoapy.NSPasteboardTypeString)

    def get_clipboard_text(self) -> str:
        text = ''
        with AutoReleasePool():
            pasteboard = NSPasteboard.generalPasteboard()

            if pasteboard.types().containsObject_(cocoapy.NSPasteboardTypeString):
                text_obj = pasteboard.stringForType_(cocoapy.NSPasteboardTypeString)
                if text_obj:
                    text = text_obj.UTF8String().decode('utf-8')

        return text


__all__ = ['CocoaWindow']



import pyglet


class Scene:

    manager: "SceneManager"

    def set_scene(self, scene_name):
        self.manager.set_scene(scene_name)

    def activate(self):
        pass

    def deactivate(self):
        pass

    def draw(self):
        pass

    def update_scene(self, dt):
        pass

    # Controller event handlers:

    def on_dpad_motion(self, controller, vector):
        pass

    def on_stick_motion(self, controller, name, vector):
        pass

    def on_trigger_motion(self, controller, value):
        pass

    def on_button_press(self, controller, button):
        pass

    def on_button_release(self, controller, button):
        pass

    # Keyboard event handlers:

    def on_key_press(self, symbol, modifiers):
        pass

    def on_key_release(self, symbol, modifiers):
        pass


class SceneManager:

    def __init__(self, window):
        self.window = window
        self.window.on_draw = self._on_draw

        self._scenes = {}
        self.current_scene = None

        # Instantiation a ControllerManager to handle hot-plugging:
        self.controller = None
        self.controller_manager = pyglet.input.ControllerManager()
        self.controller_manager.on_connect = self.on_controller_connect
        self.controller_manager.on_disconnect = self.on_controller_disconnect

        # Initialize Controller if it's connected:
        controllers = self.controller_manager.get_controllers()
        if controllers:
            self.on_controller_connect(controllers[0])

    def _on_draw(self):
        self.window.clear()
        self.current_scene.draw()

    def on_controller_connect(self, controller):
        if not self.controller:
            controller.open()
            self.controller = controller
            self.controller.push_handlers(self.current_scene)
        else:
            print(f"A Controller is already connected: {self.controller}")

    def on_controller_disconnect(self, controller):
        if self.controller == controller:
            self.controller.remove_handlers(self.current_scene)
            self.controller = None

    def add_scene(self, scene_class, *args, alias=None):
        scene_class.manager = self
        name = alias or scene_class.__name__

        scene_instance = scene_class(*args)

        self._scenes[name] = scene_instance
        self.current_scene = scene_instance
        self.window.clear()

    def set_scene(self, scene):
        assert scene in self._scenes, "Scene not found! Did you add it?"

        if self.current_scene:
            self.current_scene.deactivate()
            self.window.remove_handlers(self.current_scene)
            if self.controller:
                self.controller.remove_handlers(self.current_scene)

        self.current_scene = self._scenes[scene]
        self.current_scene.activate()
        self.window.push_handlers(self.current_scene)
        if self.controller:
            self.controller.push_handlers(self.current_scene)

    def update(self, dt):
        self.current_scene.update_scene(dt)



import pyglet

window = pyglet.window.Window(540, 500, caption="Widget Example")
batch = pyglet.graphics.Batch()
pyglet.gl.glClearColor(0.8, 0.8, 0.8, 1.0)


@window.event
def on_draw():
    window.clear()
    batch.draw()


####################################
# load resources to use for Widgets:
####################################

unpressed = pyglet.resource.image('button_unpressed.png')
pressed = pyglet.resource.image('button_pressed.png')
hover = pyglet.resource.image('button_hover.png')
bar = pyglet.resource.image('bar.png')
knob = pyglet.resource.image('knob.png')


######################################
# Create some event handler functions:
######################################

def slider_handler(widget, value):
    slider_label.text = f"Slider Value: {round(value, 1)}"


def toggle_button_handler(widget, value):
    toggle_label.text = f"Toggle Button: {value}"


def push_button_handler(widget):
    push_label.text = "Push Button: True"


def release_button_handler(widget):
    push_label.text = "Push Button: False"


def text_entry_handler(widget, text):
    text_entry_label.text = f"Text: {text}"


###############################
# Create some Widget instances:
###############################

# A Frame instance to hold all widgets, and provide special
# hashing to avoid sending all the Window events to every widget:
frame = pyglet.gui.Frame(window, order=4)


togglebutton = pyglet.gui.ToggleButton(100, 400, pressed=pressed, unpressed=unpressed, hover=hover, batch=batch)
togglebutton.set_handler('on_toggle', toggle_button_handler)
frame.add_widget(togglebutton)
toggle_label = pyglet.text.Label("Toggle Button: False", x=300, y=400, batch=batch, color=(0, 0, 0, 255))


pushbutton = pyglet.gui.PushButton(100, 300, pressed=pressed, unpressed=unpressed, hover=hover, batch=batch)
pushbutton.set_handler('on_press', push_button_handler)
pushbutton.set_handler('on_release', release_button_handler)
frame.add_widget(pushbutton)
push_label = pyglet.text.Label("Push Button: False", x=300, y=300, batch=batch, color=(0, 0, 0, 255))


slider = pyglet.gui.Slider(100, 200, bar, knob, edge=5, batch=batch)
slider.set_handler('on_change', slider_handler)
frame.add_widget(slider)
slider_label = pyglet.text.Label("Slider Value: 0.0", x=300, y=200, batch=batch, color=(0, 0, 0, 255))


# This Widget is not added to the Frame. Because it is sensitive
# to drag-and-select events falling outside the Frame's spatial hash,
# it's best to let it handle Window events directly.
text_entry = pyglet.gui.TextEntry("Enter Your Name", 100, 100, 150, batch=batch)
window.push_handlers(text_entry)
text_entry.set_handler('on_commit', text_entry_handler)
text_entry_label = pyglet.text.Label("Text: None", x=300, y=100, batch=batch, color=(0, 0, 0, 255))


pyglet.app.run()



"""
Base classes for test cases using the normal pyglet application event loop.
"""

import pytest

import pyglet
from pyglet import clock
from pyglet import gl
from pyglet.graphics import Batch
from pyglet.text.document import FormattedDocument
from pyglet.text.layout import TextLayout
from pyglet.window import Window, key

from .interactive import InteractiveFixture


@pytest.fixture
def event_loop(request):
    return EventLoopFixture(request)


class EventLoopFixture(InteractiveFixture):
    question = '\n\n(P)ass/(F)ail/(S)kip/(Q)uit?'
    key_pass = key.P
    key_fail = key.F
    key_skip = key.S
    key_quit = key.Q
    clear_color = 1, 1, 1, 1
    base_options = {
        'width': 300,
        'height': 300,
    }

    def __init__(self, request):
        super().__init__(request)
        self._request = request
        self.window = None
        self.text_batch = None
        self.text_document = None
        self.answer = None
        request.addfinalizer(self.tear_down)

    def tear_down(self):
        if self.window:
            self.window.close()
            self.window = None

    def create_window(self, **kwargs):
        combined_kwargs = {}
        combined_kwargs.update(self.base_options)
        combined_kwargs.update(kwargs)
        self.window = Window(**combined_kwargs)
        self.window.push_handlers(self)
        return self.window

    def get_document(self):
        if self.text_document is None:
            self._create_text()
        return self.text_document

    def _create_text(self):
        assert self.window is not None
        self.text_batch = Batch()
        self.text_document = FormattedDocument()
        layout = TextLayout(self.text_document, self.window.width, self.window.height,
                            multiline=True, wrap_lines=True, batch=self.text_batch)
        layout.content_valign = 'bottom'

    def add_text(self, text):
        self.get_document()
        self.text_document.insert_text(len(self.text_document.text), text)

    def ask_question(self, description=None, screenshot=True):
        """Ask a question inside the test window. By default takes a screenshot and validates
        that too."""
        if self.window is None:
            self.create_window()
        self.add_text('\n\n')
        if description:
            self.add_text(description)
        self.add_text(self.question)
        self.answer = None
        caught_exception = None
        try:
            if self.interactive:
                self.run_event_loop()
                self.handle_answer()
            else:
                self.run_event_loop(0.1)
        except Exception as ex:
            import traceback
            traceback.print_exc()
            caught_exception = ex
        finally:
            if screenshot:
                try:
                    screenshot_name = self._take_screenshot(self.window)
                    if caught_exception is None and not self.interactive:
                        self._check_screenshot(screenshot_name)
                except:
                    if not caught_exception:
                        raise
            if caught_exception:
                raise caught_exception

    def handle_answer(self):
        if self.answer is None:
            raise Exception('Did not receive valid input in question window')
        elif self.answer == self.key_fail:
            # TODO: Ask input
            pytest.fail('Tester marked test failed')
        elif self.answer == self.key_skip:
            pytest.skip('Tester marked test skipped')
        elif self.answer == self.key_quit:
            pytest.exit('Tester requested to quit')

    def ask_question_no_window(self, description=None):
        """Ask a question to verify the current test result. Uses the console or an external gui
        as no window is available."""
        super().ask_question(description)

    def run_event_loop(self, duration=None):
        if duration:
            clock.schedule_once(self.interrupt_event_loop, duration)
        pyglet.app.run()

    def interrupt_event_loop(self, *args, **kwargs):
        pyglet.app.exit()

    @staticmethod
    def schedule_once(callback, dt=.1):
        clock.schedule_once(callback, dt)

    def on_draw(self):
        self.clear()
        self.draw_text()

    def clear(self):
        gl.glClearColor(*self.clear_color)
        self.window.clear()

    def draw_text(self):
        if self.text_batch is not None:
            self.text_batch.draw()

    def on_key_press(self, symbol, modifiers):
        if symbol in (self.key_pass, self.key_fail, self.key_skip, self.key_quit):
            self.answer = symbol
            self.interrupt_event_loop()

        # Prevent handling of Escape to close the window
        return True


def test_question_pass(event_loop):
    event_loop.create_window()
    event_loop.ask_question('If you read this text, you should let the test pass.')


def test_question_fail(event_loop):
    event_loop.create_window()
    with pytest.raises(pytest.fail.Exception):
        event_loop.ask_question('Please press F to fail this test.')


def test_question_skip(event_loop):
    event_loop.create_window()
    event_loop.ask_question('Please press S to skip the rest of this test.')
    pytest.fail('You should have pressed S')



import pytest
from tests.base.interactive import InteractiveTestCase

from pyglet.gl import *
from pyglet import window
from pyglet import clock
from pyglet.window import key
import pyglet


@pytest.mark.requires_user_action
class WINDOW_MULTISAMPLE(InteractiveTestCase):
    """Test that a window can have multisample.

    A window will be opened containing two rotating squares.  Initially,
    there will be no multisampling (the edges will look "jaggy").  Press:

        * M to toggle multisampling on/off
        * S to increase samples (2, 4, 6, 8, 10, ...)
        * Shift+S to decrease samples

    Each time sample_buffers or samples is modified, the window will be recreated.
    Watch the console for success and failure messages.  If the multisample
    options are not supported, a "Failure" message will be printed and the window
    will be left as-is.

    Press ESC to end the test.
    """
    win = None
    width = 640
    height = 480

    soft_multisample = True
    multisample = False
    samples = 2

    # This test does not work on all hardware, unless rendered to texture.
    texture = pyglet.image.Texture.create(width, height, GL_TEXTURE_RECTANGLE)

    def set_window(self):
        oldwindow = self.win
        try:
            if self.multisample:
                print('Attempting samples=%d...' % self.samples, end=' ')
                config = Config(sample_buffers=1,
                                samples=self.samples,
                                double_buffer=True)
            else:
                print('Disabling multisample...', end=' ')
                config = Config(double_buffer=True)
            self.win = window.Window(self.width, self.height,
                                     vsync=True,
                                     config=config)
            self.win.switch_to()
            self.win.push_handlers(self.on_key_press)

            if self.multisample:
                if self.soft_multisample:
                    glEnable(GL_MULTISAMPLE_ARB)
                else:
                    glDisable(GL_MULTISAMPLE_ARB)

            if oldwindow:
                oldwindow.close()

            print('Success.')
        except window.NoSuchConfigException:
            print('Failed.')

    def on_key_press(self, symbol, modifiers):
        mod = 1
        if modifiers & key.MOD_SHIFT:
            mod = -1

        if symbol == key.M:
            self.multisample = not self.multisample
            self.set_window()
        if symbol == key.S:
            self.samples += 2 * mod
            self.samples = max(2, self.samples)
            self.set_window()

        # Another test: try enabling/disabling GL_MULTISAMPLE_ARB...
        # seems to have no effect if samples > 4.
        if symbol == key.N:
            self.soft_multisample = not self.soft_multisample
            if self.soft_multisample:
                print('Enabling GL_MULTISAMPLE_ARB')
                glEnable(GL_MULTISAMPLE_ARB)
            else:
                print('Disabling GL_MULTISAMPLE_ARB')
                glDisable(GL_MULTISAMPLE_ARB)

    def render(self):
        self.win.switch_to()
        size = self.height / 4
        glClear(GL_COLOR_BUFFER_BIT)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

        glLoadIdentity()
        glTranslatef(self.width/2, self.height/2, 0)
        glRotatef(self.angle, 0, 0, 1)

        glColor3f(1, 0, 0)
        glBegin(GL_QUADS)
        glVertex2f(-size, -size)
        glVertex2f(size, -size)
        glVertex2f(size, size)
        glVertex2f(-size, size)
        glEnd()

        glRotatef(-self.angle * 2, 0, 0, 1)
        glColor4f(0, 1, 0, 0.5)
        glBegin(GL_QUADS)
        glVertex2f(-size, -size)
        glVertex2f(size, -size)
        glVertex2f(size, size)
        glVertex2f(-size, size)
        glEnd()

        # Render to texture, then blit to screen:
        buffer = pyglet.image.get_buffer_manager().get_color_buffer()
        self.texture.blit_into(buffer, 0, 0, 0)
        glViewport(0, 0, self.width, self.height)
        glClearColor(0, 0, 0, 1)
        glClear(GL_COLOR_BUFFER_BIT)
        glLoadIdentity()
        glColor3f(1, 1, 1)
        self.texture.blit(0, 0, width=self.width, height=self.height)

    def test_multisample(self):
        self.set_window()
        try:
            self.angle = 0
            while not self.win.has_exit:
                dt = clock.tick()
                self.angle += dt

                self.render()
                self.win.flip()
                self.win.dispatch_events()
        finally:
            self.win.close()
        self.user_verify('Pass test?', take_screenshot=False)




"""
Tests for behavior unique to BorderedRectangle.

Most color property behavior is tested with the other _ShapeBase
subclasses in test_shapes.py
"""
import pytest

from pyglet.shapes import BorderedRectangle


@pytest.fixture
def bordered_rectangle():
    return BorderedRectangle(0, 0, 10, 50, color=(0, 0, 0, 31), border_color=(1, 1, 1))


def test_init_raises_value_error_on_conflicting_rgba_alphas():
    with pytest.raises(ValueError):
        _ = BorderedRectangle(0, 0, 10, 10, color=(1, 1, 1, 2), border_color=(255, 255, 255, 255))


@pytest.mark.parametrize(
    'color, border_color, expected_alpha', [
        ((0, 0, 0), (3, 3, 3), 255),         # 255 alpha when both colors RGB
        ((2, 2, 2, 5), (7, 7, 7), 5),        # color is RGBA, its alpha used
        ((17, 17, 17), (23, 23, 23, 0), 0),  # border_color's alpha used
        ((1, 1, 1, 1), (2, 2, 2, 1), 1)      # identical alphas are accepted
    ]
)
def test_init_sets_opacity_for_valid_color_and_border_color_alphas(color, border_color, expected_alpha):
    rect = BorderedRectangle(0, 0, 10, 50, color=color, border_color=border_color)
    assert rect.opacity == expected_alpha


def test_setting_border_color_sets_border_color_rgb_channels(bordered_rectangle, new_rgb_or_rgba_color):
    bordered_rectangle.border_color = new_rgb_or_rgba_color
    assert bordered_rectangle.border_color[:3] == new_rgb_or_rgba_color[:3]


def test_setting_border_color_to_rgb_value_does_not_change_opacity(bordered_rectangle, new_rgb_color):
    original_opacity = bordered_rectangle.opacity
    bordered_rectangle.border_color = new_rgb_color
    assert bordered_rectangle.opacity == original_opacity
    assert bordered_rectangle.color[3] == original_opacity
    assert bordered_rectangle.border_color[3] == original_opacity


def test_setting_border_color_to_rgba_value_does_not_change_fill_color_rgb_channels(bordered_rectangle, new_rgba_color):
    original_rgb_color = bordered_rectangle.color[:3]
    bordered_rectangle.border_color = new_rgba_color
    assert bordered_rectangle.color[:3] == original_rgb_color


def test_setting_border_color_to_rgba_value_sets_opacity(bordered_rectangle, new_rgba_color):
    new_opacity = new_rgba_color[3]
    bordered_rectangle.border_color = new_rgba_color
    assert bordered_rectangle.opacity == new_opacity
    assert bordered_rectangle.color[3] == new_opacity
    assert bordered_rectangle.border_color[3] == new_opacity


def test_setting_color_to_rgb_value_does_not_change_alpha_channel_on_border_color(bordered_rectangle, new_rgb_color):
    original_opacity = bordered_rectangle.opacity
    bordered_rectangle.color = new_rgb_color
    assert bordered_rectangle.border_color[3] == original_opacity


def test_setting_color_to_rgba_value_does_not_change_rgb_channels_on_border_color(bordered_rectangle, new_rgba_color):
    original_border_rgb = bordered_rectangle.border_color[:3]
    bordered_rectangle.color = new_rgba_color
    assert bordered_rectangle.border_color[:3] == original_border_rgb


def test_setting_color_to_rgba_value_sets_alpha_channel_on_border_color(bordered_rectangle, new_rgba_color):
    new_opacity = new_rgba_color[3]
    bordered_rectangle.color = new_rgba_color
    assert bordered_rectangle.border_color[3] == new_opacity


def test_setting_opacity_does_not_change_border_color_rgb_channels(bordered_rectangle):
    original_border_rgb = bordered_rectangle.border_color[:3]
    bordered_rectangle.opacity = 47
    assert bordered_rectangle.border_color[:3] == original_border_rgb



"""
Responsibilities

    Handles accumulation of debug events while playing media_player and  saves
    when sample's play ends.
"""

import time
import pickle


class BufferedLogger:
    def __init__(self, outfile):
        self.outfile = outfile
        self.log_entries = []
        self.start_wall_time = None
        # (fn, args)
        self.on_close_callback_info = None

    def init_wall_time(self):
        self.start_wall_time = time.perf_counter()

    def log(self, *args):
        self.log_entries.append(args)

    def rebased_wall_time(self):
        return time.perf_counter() - self.start_wall_time

    def close(self):
        self.save_log_entries_as_pickle()
        if self.on_close_callback_info is not None:
            fn, args = self.on_close_callback_info
            fn(self.log_entries, *args)

    def save_log_entries_as_pickle(self):
        with open(self.outfile, "wb") as f:
            pickle.dump(self.log_entries, f)

    def clear(self):
        self.log_entries = []


logger = None



#!/usr/bin/python
# $Id:$

import ctypes
import math
import optparse
import pyglet

root2 = math.sqrt(2)

def gen_dist_field(img, width, height, spread, bidirectional=True):
    sample_width = img.width // width
    sample_height = img.height // height
    max_dist = spread * root2
    dist_scale = 127. / max_dist

    # Grab image data as list of int.
    data = img.get_data('L', img.width)
    if isinstance(data, str):
        data = map(ord, data)

    # Pad input image by `spread` on each side to avoid having to check bounds
    # Add left and right sides
    data_tmp = []
    padx = [0] * spread
    pitch = img.width
    in_i = 0
    for row_i in range(img.height):
        data_tmp.extend(padx)
        data_tmp.extend(data[in_i:in_i + pitch])
        data_tmp.extend(padx)
        in_i += pitch
    pitch += spread * 2
    # Add top and bottom
    pady = [0] * (pitch * spread)
    data = pady + data_tmp + pady

    field = []
    for y in range(height):
        in_i = (spread + int((y + 0.5) * sample_height)) * pitch + \
            spread + sample_width // 2
        for x in range(width):
            colour = data[in_i]
            dist1_sq = dist2_sq = dist_sq = (max_dist + 1) ** 2
            for sy in range(-spread, spread + 1):
                row_i = in_i + sy * pitch
                for sx in range(-spread, spread + 1):
                    if data[row_i + sx] != colour:
                        dist_sq = min(dist_sq, sy * sy + sx * sx)
                        if sx * sy > 0:
                            dist1_sq = min(dist1_sq, sy * sy + sx * sx)
                        else:
                            dist2_sq = min(dist2_sq, sy * sy + sx * sx)
            dist = math.sqrt(dist_sq)
            dist1 = math.sqrt(dist1_sq)
            dist2 = math.sqrt(dist2_sq)

            if not colour:
                dist = -dist
                dist1 = -dist1
                dist2 = -dist2
            dist = int(dist * dist_scale) + 128
            dist1 = int(dist1 * dist_scale) + 128
            dist2 = int(dist2 * dist_scale) + 128
            if bidirectional:
                field.append(dist1)
                field.append(dist2)
                field.append(dist)
            else:
                field.append(255)
            field.append(dist)
            in_i += sample_width

    field = (ctypes.c_byte * len(field))(*field)

    out_img = pyglet.image.create(width, height)
    if bidirectional:
        out_img.set_data('RGBA', width * 4, field)
    else:
        out_img.set_data('LA', width * 2, field)
    return out_img

if __name__ == '__main__':
    import os
    
    parser = optparse.OptionParser()
    parser.add_option('-s', '--size', type=int, default=32)
    parser.add_option('-p', '--spread', type=int, default=128)
    parser.add_option('-o', '--output', default=None)
    parser.add_option('-b', '--bidirectional', action='store_true', 
                      default=False)
    (options, args) = parser.parse_args()

    filename = args[0]
    img = pyglet.image.load(filename)
    img2 = gen_dist_field(img, options.size, options.size, 
                          options.spread, options.bidirectional)
    if not options.output:
        base, _ = os.path.splitext(filename)
        options.output = f'{base}.df.png'
    img2.save(options.output)




"""Decode HTML into attributed text.

A subset of HTML 4.01 Transitional is implemented.  The following elements are
supported fully::

    B BLOCKQUOTE BR CENTER CODE DD DIR DL EM FONT H1 H2 H3 H4 H5 H6 I IMG KBD
    LI MENU OL P PRE Q SAMP STRONG SUB SUP TT U UL VAR

The mark (bullet or number) of a list item is separated from the body of the
list item with a tab, as the pyglet document model does not allow
out-of-stream text.  This means lists display as expected, but behave a little
oddly if edited.

No CSS styling is supported.
"""
from __future__ import annotations

import contextlib
import re
from html import entities
from html.parser import HTMLParser
from typing import TYPE_CHECKING, Any, ClassVar

import pyglet
from pyglet.text.formats import structured

if TYPE_CHECKING:
    from pyglet.image import AbstractImage
    from pyglet.resource import Location


def _hex_color(val: int) -> list[int, int, int, int]: # type: ignore reportInvalidTypeArguments
    return [(val >> 16) & 0xff, (val >> 8) & 0xff, val & 0xff, 255]


_color_names = {
    "black":    _hex_color(0x000000),
    "silver":   _hex_color(0xc0c0c0),
    "gray":     _hex_color(0x808080),
    "white":    _hex_color(0xffffff),
    "maroon":   _hex_color(0x800000),
    "red":      _hex_color(0xff0000),
    "purple":   _hex_color(0x800080),
    "fucsia":   _hex_color(0x008000),
    "green":    _hex_color(0x00ff00),
    "lime":     _hex_color(0xffff00),
    "olive":    _hex_color(0x808000),
    "yellow":   _hex_color(0xff0000),
    "navy":     _hex_color(0x000080),
    "blue":     _hex_color(0x0000ff),
    "teal":     _hex_color(0x008080),
    "aqua":     _hex_color(0x00ffff),
}


def _parse_color(value: str) -> list[int, int, int, int]: # type: ignore reportInvalidTypeArguments
    if value.startswith("#"):
        return _hex_color(int(value[1:], 16))

    try:
        return _color_names[value.lower()]
    except KeyError:
        pass

    raise ValueError


_whitespace_re = re.compile("[\u0020\u0009\u000c\u200b\r\n]+", re.DOTALL)

_metadata_elements = ["head", "title"]

_block_elements = ["p", "h1", "h2", "h3", "h4", "h5", "h6",
                   "ul", "ol", "dir", "menu",
                   "pre", "dl", "div", "center",
                   "noscript", "noframes", "blockquote", "form",
                   "isindex", "hr", "table", "fieldset", "address",
                   # Incorrect, but we treat list items as blocks:
                   "li", "dd", "dt" ]

_block_containers = ["_top_block",
                     "body", "div", "center", "object", "applet",
                     "blockquote", "ins", "del", "dd", "li", "form",
                     "fieldset", "button", "th", "td", "iframe", "noscript",
                     "noframes",
                     # Incorrect, but we treat list items as blocks:
                     "ul", "ol", "dir", "menu", "dl"]


class HTMLDecoder(HTMLParser, structured.StructuredTextDecoder):
    """Decoder for HTML documents."""
    #: Default style attributes for unstyled text in the HTML document.
    default_style: ClassVar[dict[str, Any]] = {
        "font_name": "Times New Roman",
        "font_size": 12,
        "margin_bottom": "12pt",
        "weight": "normal",
        "italic": False,
    }

    #: Map HTML font sizes to actual font sizes, in points.
    font_sizes: ClassVar[dict[int, int]] = {
        1: 8,
        2: 10,
        3: 12,
        4: 14,
        5: 18,
        6: 24,
        7: 48,
    }

    def decode_structured(self, text: str, location: Location) -> None:
        self.location = location
        self._font_size_stack = [3]
        self.list_stack.append(structured.UnorderedListBuilder({})) # type: ignore reportArgumentType
        self.strip_leading_space = True
        self.block_begin = True
        self.need_block_begin = False
        self.element_stack = ["_top_block"]
        self.in_metadata = False
        self.in_pre = False

        self.push_style("_default", self.default_style)

        self.feed(text)
        self.close()

    def get_image(self, filename: str) -> AbstractImage:
        return pyglet.image.load(filename, file=self.location.open(filename))

    def prepare_for_data(self) -> None:
        if self.need_block_begin:
            self.add_text("\n")
            self.block_begin = True
            self.need_block_begin = False

    def handle_data(self, data: str) -> None:
        if self.in_metadata:
            return

        if self.in_pre:
            self.add_text(data)
        else:
            data = _whitespace_re.sub(" ", data)
            if data.strip():
                self.prepare_for_data()
                if self.block_begin or self.strip_leading_space:
                    data = data.lstrip()
                    self.block_begin = False
                self.add_text(data)
            self.strip_leading_space = data.endswith(" ")

    def handle_starttag(self, tag: str, case_attrs: dict[str, Any]) -> None:
        if self.in_metadata:
            return

        element = tag.lower()
        attrs = {}
        for key, value in case_attrs:
            attrs[key.lower()] = value

        if element in _metadata_elements:
            self.in_metadata = True
        elif element in _block_elements:
            # Pop off elements until we get to a block container.
            while self.element_stack[-1] not in _block_containers:
                self.handle_endtag(self.element_stack[-1])
            if not self.block_begin:
                self.add_text("\n")
                self.block_begin = True
                self.need_block_begin = False
        self.element_stack.append(element)

        style = {}
        if element in ("b", "strong"):
            style["weight"] = "bold"
        elif element in ("i", "em", "var"):
            style["italic"] = True
        elif element in ("tt", "code", "samp", "kbd"):
            style["font_name"] = "Courier New"
        elif element == "u":
            color = self.current_style.get("color")
            if color is None:
                color = [0, 0, 0, 255]
            style["underline"] = color
        elif element == "font":
            if "face" in attrs:
                style["font_name"] = attrs["face"].split(",")
            if "size" in attrs:
                size = attrs["size"]
                try:
                    if size.startswith("+"):
                        size = self._font_size_stack[-1] + int(size[1:])
                    elif size.startswith("-"):
                        size = self._font_size_stack[-1] - int(size[1:])
                    else:
                        size = int(size)
                except ValueError:
                    size = 3
                self._font_size_stack.append(size)
                if size in self.font_sizes:
                    style["font_size"] = self.font_sizes.get(size, 3)
            elif "real_size" in attrs:
                size = int(attrs["real_size"])
                self._font_size_stack.append(size)
                style["font_size"] = size
            else:
                self._font_size_stack.append(self._font_size_stack[-1])
            if "color" in attrs:
                with contextlib.suppress(ValueError):
                    style["color"] = _parse_color(attrs["color"])
        elif element == "sup":
            size = self._font_size_stack[-1] - 1
            style["font_size"] = self.font_sizes.get(size, 1)
            style["baseline"] = "3pt"
        elif element == "sub":
            size = self._font_size_stack[-1] - 1
            style["font_size"] = self.font_sizes.get(size, 1)
            style["baseline"] = "-3pt"
        elif element == "h1":
            style["font_size"] = 24
            style["weight"] = "bold"
            style["align"] = "center"
        elif element == "h2":
            style["font_size"] = 18
            style["weight"] = "bold"
        elif element == "h3":
            style["font_size"] = 16
            style["weight"] = "bold"
        elif element == "h4":
            style["font_size"] = 14
            style["weight"] = "bold"
        elif element == "h5":
            style["font_size"] = 12
            style["weight"] = "bold"
        elif element == "h6":
            style["font_size"] = 12
            style["italic"] = True
        elif element == "br":
            self.add_text("\u2028")
            self.strip_leading_space = True
        elif element == "p":
            if attrs.get("align") in ("left", "center", "right"):
                style["align"] = attrs["align"]
        elif element == "center":
            style["align"] = "center"
        elif element == "pre":
            style["font_name"] = "Courier New"
            style["margin_bottom"] = 0
            self.in_pre = True
        elif element == "blockquote":
            left_margin = self.current_style.get("margin_left") or 0
            right_margin = self.current_style.get("margin_right") or 0
            style["margin_left"] = left_margin + 60
            style["margin_right"] = right_margin + 60
        elif element == "q":
            self.handle_data("\u201c")
        elif element == "ol":
            try:
                start = int(attrs.get("start", 1))
            except ValueError:
                start = 1
            fmt = attrs.get("type", "1") + "."
            builder = structured.OrderedListBuilder(start, fmt)
            builder.begin(self, style)
            self.list_stack.append(builder)
        elif element in ("ul", "dir", "menu"):
            type_ = attrs.get("type", "disc").lower()
            if type_ == "circle":
                mark = "\u25cb"
            elif type_ == "square":
                mark = "\u25a1"
            else:
                mark = "\u25cf"
            builder = structured.UnorderedListBuilder(mark)
            builder.begin(self, style)
            self.list_stack.append(builder)
        elif element == "li":
            self.list_stack[-1].item(self, style)
            self.strip_leading_space = True
        elif element == "dl":
            style["margin_bottom"] = 0
        elif element == "dd":
            left_margin = self.current_style.get("margin_left") or 0
            style["margin_left"] = left_margin + 30
        elif element == "img":
            image = self.get_image(attrs.get("src")) # type: ignore reportArgumentType
            if image:
                width = attrs.get("width")
                if width:
                    width = int(width)
                height = attrs.get("height")
                if height:
                    height = int(height)
                self.prepare_for_data()
                self.add_element(structured.ImageElement(image, width, height))
                self.strip_leading_space = False

        self.push_style(element, style)

    def handle_endtag(self, tag: str) -> None:
        element = tag.lower()
        if element not in self.element_stack:
            return

        self.pop_style(element)
        while self.element_stack.pop() != element:
            pass

        if element in _metadata_elements:
            self.in_metadata = False
        elif element in _block_elements:
            self.block_begin = False
            self.need_block_begin = True

        if element == "font" and len(self._font_size_stack) > 1:
            self._font_size_stack.pop()
        elif element == "pre":
            self.in_pre = False
        elif element == "q":
            self.handle_data("\u201d")
        elif element in ("ul", "ol") and len(self.list_stack) > 1:
            self.list_stack.pop()

    def handle_entityref(self, name: str) -> None:
        if name in entities.name2codepoint:
            self.handle_data(chr(entities.name2codepoint[name]))

    def handle_charref(self, name: str) -> None:
        name = name.lower()
        try:
            if name.startswith("x"):
                self.handle_data(chr(int(name[1:], 16)))
            else:
                self.handle_data(chr(int(name)))
        except ValueError:
            pass



#!/usr/bin/python
# $Id:$

"""Display font information.

Usage::

    inspect_font.py <filename> [<filename> ...]
"""

import sys
from pyglet.font import ttf


def inspect_font(filename):
    try:
        info = ttf.TruetypeInfo(filename)
        print('{0}:'.format(filename))
        print(info.get_name('family'))
        print('bold=%r' % info.is_bold())
        print('italic=%r' % info.is_italic())
    except:
        print("%s could not be identified.  It is probably not a TrueType or "
              "OpenType font.  However, pyglet may still be able to load it "
              "on some platforms." % filename)


if __name__ == '__main__':
    if len(sys.argv) < 2:
        print(__doc__)
    for file_name in sys.argv[1:]:
        inspect_font(file_name)



import pyglet
import random
from . import physicalobject, resources, util


def player_lives(num_icons, batch=None):
    """Generate sprites for player life icons"""
    player_lives = []
    for i in range(num_icons):
        new_sprite = pyglet.sprite.Sprite(img=resources.player_image,
                                          x=785 - i * 30, y=585,
                                          batch=batch)
        new_sprite.scale = 0.5
        player_lives.append(new_sprite)
    return player_lives


def asteroids(num_asteroids, player_position, batch=None):
    """Generate asteroid objects with random positions and velocities, not close to the player"""
    asteroids = []
    for i in range(num_asteroids):
        asteroid_x, asteroid_y, _ = player_position
        while util.distance((asteroid_x, asteroid_y), player_position) < 100:
            asteroid_x = random.randint(0, 800)
            asteroid_y = random.randint(0, 600)
        new_asteroid = physicalobject.PhysicalObject(img=resources.asteroid_image,
                                                     x=asteroid_x, y=asteroid_y,
                                                     batch=batch)
        new_asteroid.rotation = random.randint(0, 360)
        new_asteroid.velocity_x, new_asteroid.velocity_y = random.random() * 40, random.random() * 40
        asteroids.append(new_asteroid)
    return asteroids



from __future__ import annotations

import json
import struct

from array import array
from urllib.request import urlopen

import pyglet

from pyglet.gl import GL_BYTE, GL_UNSIGNED_BYTE, GL_SHORT, GL_UNSIGNED_SHORT, GL_FLOAT, GL_DOUBLE
from pyglet.gl import GL_INT, GL_UNSIGNED_INT, GL_ELEMENT_ARRAY_BUFFER, GL_ARRAY_BUFFER
from pyglet.gl import GL_REPEAT

from . import ModelDecodeException, ModelDecoder
from .base import Scene
from .base import PBRMaterial
from .base import Camera as BaseCamera
from .base import Attribute as BaseAttribute
from .base import Primitive as BasePrimitive
from .base import Node as BaseNode
from .base import Mesh as BaseMesh


_array_types = {
    GL_BYTE: 'b',
    GL_UNSIGNED_BYTE: 'B',
    GL_SHORT: 'h',
    GL_UNSIGNED_SHORT: 'H',
    GL_INT: 'l',
    GL_UNSIGNED_INT: 'L',
    GL_FLOAT: 'f',
    GL_DOUBLE: 'd',
}

_gl_type_sizes = {
    GL_BYTE: 1,
    GL_UNSIGNED_BYTE: 1,
    GL_SHORT: 2,
    GL_UNSIGNED_SHORT: 2,
    GL_INT: 4,
    GL_UNSIGNED_INT: 4,
    GL_FLOAT: 4,
    GL_DOUBLE: 8,
}

_accessor_type_counts = {
    "SCALAR": 1,
    "VEC2": 2,
    "VEC3": 3,
    "VEC4": 4,
    "MAT2": 4,
    "MAT3": 9,
    "MAT4": 16
}

_targets = {
    GL_ELEMENT_ARRAY_BUFFER: "ELEMENT_ARRAY_BUFFER",
    GL_ARRAY_BUFFER: "ARRAY_BUFFER",
}


class Buffer:
    """Abstraction over unstructured bytes."""
    def __init__(self, data):
        self.length = data['byteLength']
        self._uri = data['uri']

        if self._uri.startswith('data'):
            self._response = urlopen(self._uri)
            self._file = self._response.file
        else:
            self._file = pyglet.resource.file(self._uri, 'rb')

    def read(self, offset, nbytes) -> bytes:
        self._file.seek(offset)
        return self._file.read(nbytes)

    def __del__(self):
        try:
            self._file.close()
        except AttributeError:
            pass

    def __repr__(self):
        return f"{self.__class__.__name__}(length={self.length})"


class BufferView:
    """View over a section of a Buffer, with optional stride."""
    def __init__(self, data, owner):
        self._buffer_index = data.get('buffer')
        self.buffer = owner.buffers[self._buffer_index]

        self.offset = data.get('byteOffset', 0)
        self.length = data.get('byteLength')
        self.stride = data.get('byteStride', 0)
        self.target = data.get('target')
        self.target_name = _targets.get(self.target)

    def read(self, read_offset: int, byte_length: int, count: int) -> bytes:
        offset = self.offset + read_offset
        bytestring = b""

        for _ in range(count):
            bytestring += self.buffer.read(offset, byte_length)
            offset += self.stride or byte_length

        return bytestring

    def __repr__(self):
        return f"{self.__class__.__name__}(buffer={self._buffer_index}, target={self.target_name})"


class Accessor:
    def __init__(self, data, owner):
        self._buffer_view_index = data.get('bufferView')
        self.buffer_view = owner.buffer_views[self._buffer_view_index]

        self.byte_offset = data.get('byteOffset', 0)
        self.component_type = data.get('componentType')     # GL_FLOAT, GL_INT, etc
        self.type = data.get('type')                        # VEC3, MAT4, etc.
        self.count = data.get('count')                      # count of self.type
        self.max = data.get('max')
        self.min = data.get('min')

        # This is a 'sparse' accessor:
        self.sparse = data.get('sparse')
        if self.sparse:
            raise NotImplementedError("Not yet implemented")

        # The Python format type:
        self.fmt = _array_types[self.component_type]

        # The byte size of the `GL type` multiplied by the length of the GLSL `data type`.
        # For example: a GL_FLOAT is 4 bytes and a VEC3 has 3 values, so 4 * 3 = 12 bytes
        self._byte_length = _gl_type_sizes[self.component_type] * _accessor_type_counts[self.type]

    def read(self) -> bytes:
        return self.buffer_view.read(self.byte_offset, self._byte_length, self.count)
        # readbytes = self.buffer_view.read(self.byte_offset, self._byte_length, self.count)
        # assert self._byte_length * self.count == len(readbytes), "insufficient bytes read"
        # return readbytes

    def as_array(self):
        return array(self.fmt, self.read())

    def __repr__(self):
        return f"{self.__class__.__name__}(buffer_view={self._buffer_view_index})"


class Attribute(BaseAttribute):
    def __init__(self, name, index, owner):
        accessor = owner.accessors[index]
        self.target = accessor.buffer_view.target
        self.target_name = accessor.buffer_view.target_name
        super().__init__(name, accessor.fmt, accessor.type, accessor.count, accessor.as_array())


class Primitive(BasePrimitive):
    def __init__(self, data, owner):
        attributes = [Attribute(name, index, owner) for name, index in data.get('attributes').items()]

        indices_index = data.get('indices')
        indices_accessor = owner.accessors[indices_index] if indices_index is not None else None
        indices = indices_accessor.as_array() if indices_accessor else None

        mode = data.get('mode', 4)     # defaults to TRIANGLES

        material_index = data.get('material')
        material = owner.materials[material_index] if material_index is not None else None

        super().__init__(attributes, indices, mode, material)


class Material(PBRMaterial):
    def __init__(self, data):
        self.name = data.get('name')
        # self.extensions = data.get('extensions')
        # self.extras = data.get('extras')

        # TODO: parse this:
        self.pbr_metallic_roughness = data.get('pbrMetallicRoughness')

        self.normal_texture = data.get('normalTexture')
        self.occlusion_texture = data.get('occlusionTexture')
        self.emissive_texture = data.get('emissiveTexture')
        self.base_color_texture = data.get('baseColorTexture')

        self.emissive_factor = data.get('emissiveFactor', (0.0, 0.0, 0.0))
        self.alpha_mode = data.get('alphaMode', 'OPAQUE')   # Any of: OPAQUE, MASK, BLEND
        self.alpha_cutoff = data.get('alphaCutoff', 0.5)
        self.double_sided = data.get('doubleSided', False)

        # TODO: finish this
        # super().__init__(name, )


class Texture:
    def __init__(self, data, owner):
        self.name = data.get('name')
        # self.extensions = data.get('extensions')
        # self.extras = data.get('extras')

        # TODO: verify how this works. Default sampler?
        self._sampler_index = data.get('sampler')
        if self._sampler_index:
            self.sampler = owner.samplers[self._sampler_index]
        else:
            self.sampler = Sampler({})
        self.source = data.get('source')            # technically NOT required
        self.image = owner.images[self.source]

        # Aliases
        self.min_filter = self.sampler.min_filter
        self.mag_filter = self.sampler.mag_filter
        self.wrap_s = self.sampler.wrap_s
        self.wrap_t = self.sampler.wrap_t


class Sampler:
    def __init__(self, data):
        # TODO: make objects for min/mag filter objects
        self.name = data.get('name')
        self.min_filter = data.get('minFilter')
        self.mag_filter = data.get('magFilter')
        self.wrap_s = data.get('wrapS', GL_REPEAT)
        self.wrap_t = data.get('wrapT', GL_REPEAT)
        # self.extensions = data.get('extensions')
        # self.extras = data.get('extras')


class Image:
    def __init__(self, data, owner):
        self.uri = data.get('uri')
        self._buffer_view_index = data.get('bufferView')
        self.buffer_view = owner.buffer_views[self._buffer_view_index] if self._buffer_view_index else None
        self.mime_type = data.get('mimeType')
        self.name = data.get('name')
        self.extensions = data.get('extensions')
        self.extras = data.get('extras')

    def read(self):
        # TODO: load from either URI or bufferview
        # if self.uri:
        #     return
        # else:
        raise NotImplementedError


class Camera(BaseCamera):
    def __init__(self, camera_type, data):
        aspect_ratio = data.get('aspectRatio')  # Not required
        yfov = data.get('yfov')
        # Orthographic
        xmag = data.get('xmag')
        ymag = data.get('ymag')
        # Shared
        zfar = data.get('zfar')                 # Not required for Perspective
        znear = data.get('znear')
        super().__init__(camera_type, aspect_ratio, yfov, xmag, ymag, zfar, znear)


class Mesh:
    def __init__(self, data, owner):
        self.data = data
        self.primitives = [Primitive(primitive_data, owner) for primitive_data in data.get('primitives')]

    def __repr__(self):
        return f"{self.__class__.__name__}(primitives={len(self.primitives)})"


class Node:
    def __init__(self, data, owner):
        self.data = data
        self._owner = owner
        self._child_indices = self.data.get('children', [])

        _mesh_index = data.get('mesh')
        self.mesh = owner.meshes[_mesh_index] if _mesh_index is not None else None

        self.matrix = data.get('matrix')            # Mat4
        self.translation = data.get('translation')  # Vec3
        self.rotation = data.get('rotation')        # Quaternion
        self.scale = data.get('scale')              # Vec3

        # TODO: handle global and local transforms:
        # https://github.com/KhronosGroup/glTF-Tutorials/blob/master/gltfTutorial/gltfTutorial_004_ScenesNodes.md

    @property
    def children(self):
        return [self._owner.nodes[i] for i in self._child_indices]

    def __iter__(self):
        yield self
        for child in self.children:
            yield child

    def __repr__(self):
        return f"{self.__class__.__name__}(mesh={self.mesh}, children={self._child_indices})"


class GLTF:
    def __init__(self, gltf_data: dict, binary_buffer: bytes | None = None):
        self._gltf_data = gltf_data
        self.version = self._gltf_data['asset']['version']
        self.generator = self._gltf_data['asset'].get('generator', 'unknown')

        self.buffers = [Buffer(data=data) for data in gltf_data['buffers']]
        self.buffer_views = [BufferView(data=data, owner=self) for data in gltf_data['bufferViews']]
        self.accessors = [Accessor(data=data, owner=self) for data in gltf_data['accessors']]

        if binary_buffer:
            # TODO: test this, and think of a better way to do it
            self.buffers[0]._file = binary_buffer

        self.images = [Image(data=data, owner=self) for data in gltf_data.get('images', [])]
        self.samplers = [Sampler(data=data) for data in gltf_data.get('samplers', [])]
        self.textures = [Texture(data=data, owner=self) for data in gltf_data.get('textures', [])]
        self.materials = [Material(data) for data in gltf_data.get('materials', [])]

        print(self.images)
        print(self.samplers)
        print(self.textures)

        self.meshes = [Mesh(data=data, owner=self) for data in gltf_data['meshes']]
        self.nodes = [Node(data=data, owner=self) for data in gltf_data['nodes']]

        self.cameras = [Camera(cam['type'], cam[cam['type']]) for cam in gltf_data.get('cameras', [])]

        self.scenes = [Scene(nodes=[self.nodes[i] for i in data['nodes']]) for data in gltf_data['scenes']]
        self.default_scene = self.scenes[gltf_data.get('scene', 0)]

    def __repr__(self):
        return f"{self.__class__.__name__}(scenes={len(self.scenes)}, meshes={len(self.meshes)})"


def load_gltf(filename, file=None) -> GLTF:

    try:
        if file is None:
            file = pyglet.resource.file(filename, 'r')
        elif file.mode != 'r':
            file.close()
            file = pyglet.resource.file(filename, 'r')
    except pyglet.resource.ResourceNotFoundException:
        raise ModelDecodeException

    if filename.endswith('glb'):
        # Check header
        magic = file.read(4)
        if magic != b"glTF":
            raise ModelDecodeException(f"Invalid header for .glb file: {magic}")

        version = struct.unpack("<I", file.read(4))[0]
        if version != 2:
            raise ModelDecodeException(f"Unsupported glTF version: {version}")

        # Total file size including headers
        _ = struct.unpack("<I", file.read(4))[0]  # noqa

        # Chunk 0 - json
        chunk_0_length = struct.unpack("<I", file.read(4))[0]
        chunk_0_type = file.read(4)
        if chunk_0_type != b"JSON":
            raise ModelDecodeException(f"glTF 'chunk 0 type' is not 'JSON': {chunk_0_type}")

        json_meta = file.read(chunk_0_length).decode()

        # chunk 1 - binary buffer
        chunk_1_length = struct.unpack("<I", file.read(4))[0]
        chunk_1_type = file.read(4)
        if chunk_1_type != b"BIN\x00":
            raise ModelDecodeException(f"glTF 'chunk 1 type' is not 'BIN': {chunk_0_type}")

        binary_buffer = file.read(chunk_1_length)

    else:
        json_meta = file.read()
        binary_buffer = None

    try:
        gltf_data = json.loads(json_meta)
    except json.JSONDecodeError as e:
        raise ModelDecodeException(f"Json error. Does not appear to be a valid glTF file. {e}")
    finally:
        file.close()

    if 'asset' not in gltf_data:
        raise ModelDecodeException("Not a valid glTF file: 'asset' property not found.")
    else:
        if float(gltf_data['asset'].get('version', 0)) < 2.0:
            raise ModelDecodeException('Only glTF 2.0+ models are supported')

    return GLTF(gltf_data=gltf_data, binary_buffer=binary_buffer)


###################################################
#   Decoder definitions start here:
###################################################

class GLTFModelDecoder(ModelDecoder):
    def get_file_extensions(self):
        return ['.gltf', '.glb']

    def decode(self, filename, file):
        gltf = load_gltf(filename, file)
        return gltf.scenes[0]


def get_decoders():
    return [GLTFModelDecoder()]


def get_encoders():
    return []



"""An example of sorting sprites in the Z axis based on the Y axis, and a standard projection."""
import pyglet
from pyglet.gl import *
import random
import math

# Standard projection Z is 0 to 255. Keep window within that.
# You will have to change window projection if you wish to go beyond this.
window = pyglet.window.Window(height=250)

fragment_source = """#version 150 core
    in vec4 vertex_colors;
    in vec3 texture_coords;
    out vec4 final_colors;

    uniform sampler2D sprite_texture;

    void main()
    {
        final_colors = texture(sprite_texture, texture_coords.xy) * vertex_colors;
        
        // No GL_ALPHA_TEST in core, use shader to discard.
        if(final_colors.a < 0.01){
            discard;
        }
    }
"""


class DepthSpriteGroup(pyglet.sprite.SpriteGroup):
    def set_state(self):
        self.program.use()

        glActiveTexture(GL_TEXTURE0)
        glBindTexture(self.texture.target, self.texture.id)

        glEnable(GL_BLEND)
        glBlendFunc(self.blend_src, self.blend_dest)

        glEnable(GL_DEPTH_TEST)
        glDepthFunc(GL_LESS)

    def unset_state(self):
        glDisable(GL_BLEND)
        glDisable(GL_DEPTH_TEST)
        self.program.stop()


class DepthSprite(pyglet.sprite.Sprite):
    group_class = DepthSpriteGroup


# Set example resource path.
pyglet.resource.path = ['../resources']
pyglet.resource.reindex()

image = pyglet.resource.image("pyglet.png")
batch = pyglet.graphics.Batch()

sprites = []

# Reuse vertex source and create new shader with alpha testing.
vertex_shader = pyglet.graphics.shader.Shader(pyglet.sprite.vertex_source, "vertex")
fragment_shader = pyglet.graphics.shader.Shader(fragment_source, "fragment")
depth_shader = pyglet.graphics.shader.ShaderProgram(vertex_shader, fragment_shader)


def make_sprite(zvalue):
    sprite = DepthSprite(image, x=0, y=0, z=zvalue, batch=batch, program=depth_shader)
    # Random color multiplier.
    sprite.color = (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))
    # Add sprites to keep in memory, like a list. Otherwise, they will get GC'd when out of scope.
    sprites.append(sprite)


sprite_count = 10
for i in range(sprite_count):
    make_sprite(i)


global elapsed
elapsed = 0


def update(dt):
    global elapsed
    for i, sprite in enumerate(sprites):
        sprite.update(x=i * .75 * (image.width - 15),
                      y=.25 * window.height * (1 + math.cos(2 * elapsed + i * math.pi / sprite_count)))
        sprite.z = sprite.y
    elapsed += dt


pyglet.clock.schedule_interval(update, 1 / 60.0)


@window.event
def on_draw():
    window.clear()
    batch.draw()


pyglet.app.run()



from __future__ import annotations

import time
import threading

from typing import TYPE_CHECKING, Set

import pyglet

from pyglet.util import debug_print

if TYPE_CHECKING:
    from pyglet.media.drivers.base import AbstractAudioPlayer


_debug = debug_print('debug_media')


class PlayerWorkerThread(threading.Thread):
    """Worker thread for refilling audio players.

    This thread manages calling the ``work`` method on low level AudioPlayer
    instances (not the high level Player instances). This allows the players
    to keep their buffers filled (and perform event dispatching tasks), but
    does not block the main thread.

    This thread will sleep for a small period between updates, but provides a
    :py:meth:`~notify` method to allow waking it immediately. A :py:meth:`~stop`
    method is provided to terminate the thread, but under normal operation it
    will exit cleanly on interpreter shutdown.
    """

    # Run every 20ms; accurate enough for event dispatching while not hogging too much
    # time updating the players
    _nap_time = 0.020

    def __init__(self) -> None:
        super().__init__(daemon=True)

        self._rest_event = threading.Event()
        # A lock that should be held as long as consistency of `self.players` is required.
        self._operation_lock = threading.Lock()
        self._stopped = False
        self.players: Set[AbstractAudioPlayer] = set()

    def run(self) -> None:
        if pyglet.options['debug_trace']:
            pyglet._install_trace()

        sleep_time = None

        while True:
            assert _debug(f"PlayerWorkerThread.run: Going to sleep "
                          f"{'indefinitely; no active players' if sleep_time is None else f'for {sleep_time}'}")
            self._rest_event.wait(sleep_time)
            self._rest_event.clear()

            assert _debug(f'PlayerWorkerThread.run: woke up @{time.time()}')
            if self._stopped:
                break

            with self._operation_lock:
                if self.players:
                    sleep_time = self._nap_time
                    for player in self.players:
                        player.work()
                else:
                    # sleep until a player is added
                    sleep_time = None

        assert _debug(f'PlayerWorkerThread.run: exiting')

    def stop(self) -> None:
        """Stop the thread and wait for it to terminate.

        The ``stop`` instance variable is set to ``True`` and the rest event
        is set.  It is the responsibility of the ``run`` method to check
        the value of ``_stopped`` after each sleep or wait and to return if
        set.
        """
        assert _debug('PlayerWorkerThread.stop()')
        self._stopped = True
        self._rest_event.set()
        try:
            self.join()
        except RuntimeError:
            # Ignore on unclean shutdown
            pass

    def notify(self) -> None:
        """Interrupt the current sleep operation.

        If the thread is currently sleeping, it will be woken immediately
        instead of waiting the full duration of the timeout.
        If the thread is not sleeping, it will run again as soon as it is
        done with its operation.
        """
        assert _debug('PlayerWorkerThread.notify()')
        self._rest_event.set()

    def add(self, player: AbstractAudioPlayer) -> None:
        """Add a player to the PlayerWorkerThread, and call :py:meth:`~notify`.

        When a player is added, it's ``work`` method will be called regularly.

        .. note:: Do not call this method from within the thread, as it will deadlock.
        """
        assert player is not None
        assert _debug('PlayerWorkerThread: player added')

        with self._operation_lock:
            self.players.add(player)

        self.notify()

    def remove(self, player: AbstractAudioPlayer) -> None:
        """Remove a player from the PlayerWorkerThread.

        This call has no effect if the player does not exist.

        .. note:: Do not call this method from within the thread, as it will deadlock.
        """
        assert _debug('PlayerWorkerThread: player removed')

        if player in self.players:
            with self._operation_lock:
                self.players.remove(player)



"""Print the details of all available input devices to stdout.
"""
import pyglet

window = pyglet.window.Window()
devices = pyglet.input.get_devices()


def watch_control(device, control):
    @control.event
    def on_change(value):
        print(f'{device!r}: {control!r}.on_change({value!r})')

    if isinstance(control, pyglet.input.base.Button):
        @control.event
        def on_press():
            print(f"{device!r}: {control!r}.on_press()")

        @control.event
        def on_release():
            print(f'{device!r}: {control!r}.on_release()')


print('Devices:')
for device in devices:
    print('  ', device.name, end=' ')
    try:
        device.open(window=window)
        print('OK')

        for control in device.get_controls():
            print('    ', control.name)
            watch_control(device, control)

    except pyglet.input.DeviceException:
        print('Fail')

pyglet.app.run()



# Windows Vista: need to call SetProcessDPIAware?  May affect GDI+ calls as well as font.
from __future__ import annotations

import ctypes
import math
import warnings
from typing import TYPE_CHECKING, ClassVar, Sequence

import pyglet
import pyglet.image
from pyglet.font import base
from pyglet.image.codecs.gdiplus import BitmapData, ImageLockModeRead, PixelFormat32bppARGB, Rect, gdiplus
from pyglet.libs.win32 import _gdi32 as gdi32
from pyglet.libs.win32 import _user32 as user32
from pyglet.libs.win32.constants import ANTIALIASED_QUALITY, FW_BOLD, FW_NORMAL
from pyglet.libs.win32.context_managers import device_context
from pyglet.libs.win32.types import ABC, BYTE, LOGFONTW, TEXTMETRIC

if TYPE_CHECKING:
    from pyglet.font.base import Glyph


DriverStringOptionsCmapLookup = 1
DriverStringOptionsRealizedAdvance = 4
TextRenderingHintSingleBitPerPixelGridFit = 1
TextRenderingHintAntiAlias = 4
TextRenderingHintAntiAliasGridFit = 3


FontStyleBold = 1
FontStyleItalic = 2
UnitPixel = 2
UnitPoint = 3

StringFormatFlagsDirectionRightToLeft = 0x00000001
StringFormatFlagsDirectionVertical = 0x00000002
StringFormatFlagsNoFitBlackBox = 0x00000004
StringFormatFlagsDisplayFormatControl = 0x00000020
StringFormatFlagsNoFontFallback = 0x00000400
StringFormatFlagsMeasureTrailingSpaces = 0x00000800
StringFormatFlagsNoWrap = 0x00001000
StringFormatFlagsLineLimit = 0x00002000
StringFormatFlagsNoClip = 0x00004000

FontFamilyNotFound = 14


_debug_font = pyglet.options["debug_font"]


class Rectf(ctypes.Structure):
    _fields_ = [
        ("x", ctypes.c_float),
        ("y", ctypes.c_float),
        ("width", ctypes.c_float),
        ("height", ctypes.c_float),
    ]


_text_quality = {
    True: TextRenderingHintAntiAliasGridFit,
    False: TextRenderingHintSingleBitPerPixelGridFit,
}

class GDIPlusGlyphRenderer(base.GlyphRenderer):

    def __init__(self, font: GDIPlusFont) -> None:
        self._bitmap = None
        self._dc = None
        self._bitmap_rect = None
        super().__init__(font)
        self.font = font

        # Pessimistically round up width and height to 4 byte alignment
        width = font.max_glyph_width
        height = font.ascent - font.descent
        width = (width | 0x3) + 1
        height = (height | 0x3) + 1
        self._create_bitmap(width, height)

        gdi32.SelectObject(self._dc, self.font.hfont)

    def __del__(self) -> None:
        try:
            if self._matrix:
                gdiplus.GdipDeleteMatrix(self._matrix)
            if self._brush:
                gdiplus.GdipDeleteBrush(self._brush)
            if self._graphics:
                gdiplus.GdipDeleteGraphics(self._graphics)
            if self._bitmap:
                gdiplus.GdipDisposeImage(self._bitmap)
            if self._dc:
                user32.ReleaseDC(0, self._dc)
        except Exception:  # noqa: S110, BLE001
            pass

    def _create_bitmap(self, width: int, height: int) -> None:
        self._data = (BYTE * (4 * width * height))()
        self._bitmap = ctypes.c_void_p()
        self._format = PixelFormat32bppARGB
        gdiplus.GdipCreateBitmapFromScan0(width, height, width * 4,
            self._format, self._data, ctypes.byref(self._bitmap))

        self._graphics = ctypes.c_void_p()
        gdiplus.GdipGetImageGraphicsContext(self._bitmap,
            ctypes.byref(self._graphics))
        gdiplus.GdipSetPageUnit(self._graphics, UnitPixel)

        self._dc = user32.GetDC(0)
        gdi32.SelectObject(self._dc, self.font.hfont)

        gdiplus.GdipSetTextRenderingHint(self._graphics,
            _text_quality[pyglet.options.text_antialiasing])

        self._brush = ctypes.c_void_p()
        gdiplus.GdipCreateSolidFill(0xffffffff, ctypes.byref(self._brush))

        self._matrix = ctypes.c_void_p()
        gdiplus.GdipCreateMatrix(ctypes.byref(self._matrix))

        self._flags = (DriverStringOptionsCmapLookup |
                       DriverStringOptionsRealizedAdvance)

        self._rect = Rect(0, 0, width, height)

        self._bitmap_height = height

    def render(self, text: str) -> Glyph:
        ch = ctypes.create_unicode_buffer(text)
        len_ch = len(text)

        # Layout rectangle; not clipped against so not terribly important.
        width = 10000
        height = self._bitmap_height
        rect = Rectf(0, self._bitmap_height
                        - self.font.ascent + self.font.descent,
                     width, height)

        # Set up GenericTypographic with 1 character measure range
        generic = ctypes.c_void_p()
        gdiplus.GdipStringFormatGetGenericTypographic(ctypes.byref(generic))
        fmt = ctypes.c_void_p()
        gdiplus.GdipCloneStringFormat(generic, ctypes.byref(fmt))
        gdiplus.GdipDeleteStringFormat(generic)

        # --- Measure advance

        # XXX HACK HACK HACK
        # Windows GDI+ is a filthy broken toy.  No way to measure the bounding
        # box of a string, or to obtain LSB.  What a joke.
        #
        # For historical note, GDI cannot be used because it cannot composite
        # into a bitmap with alpha.
        #
        # It looks like MS have abandoned GDI and GDI+ and are finally
        # supporting accurate text measurement with alpha composition in .NET
        # 2.0 (WinForms) via the TextRenderer class; this has no C interface
        # though, so we're entirely screwed.
        #
        # So anyway, we first try to get the width with GdipMeasureString.
        # Then if it's a TrueType font, we use GetCharABCWidthsW to get the
        # correct LSB. If it's a negative LSB, we move the layoutRect `rect`
        # to the right so that the whole glyph is rendered on the surface.
        # For positive LSB, we let the renderer render the correct white
        # space and we don't pass the LSB info to the Glyph.set_bearings

        bbox = Rectf()
        flags = (StringFormatFlagsMeasureTrailingSpaces |
                 StringFormatFlagsNoClip |
                 StringFormatFlagsNoFitBlackBox)
        gdiplus.GdipSetStringFormatFlags(fmt, flags)
        gdiplus.GdipMeasureString(self._graphics,
                                  ch,
                                  len_ch,
                                  self.font._gdipfont,
                                  ctypes.byref(rect),
                                  fmt,
                                  ctypes.byref(bbox),
                                  None,
                                  None)

        # We only care about the advance from this whole thing.
        advance = int(math.ceil(bbox.width))

        # GDI functions only work for a single character so we transform
        # grapheme \r\n into \r
        if text == "\r\n":
            text = "\r"

        # XXX END HACK HACK HACK

        abc = ABC()
        width = 0
        lsb = 0
        ttf_font = True
        # Use GDI to get code points for the text passed. This is almost always 1.
        # For special unicode characters it may be comprised of 2+ codepoints. Get the width/lsb of each.
        # Function only works on TTF fonts.
        for codepoint in [ord(c) for c in text]:
            if gdi32.GetCharABCWidthsW(self._dc, codepoint, codepoint, ctypes.byref(abc)):
                lsb += abc.abcA
                width += abc.abcB

                if lsb < 0:
                    # Negative LSB: we shift the layout rect to the right
                    # Otherwise we will cut the left part of the glyph
                    rect.x = -lsb
                    width -= lsb
                else:
                    width += lsb
            else:
                ttf_font = False
                break

        # Almost always a TTF font. Haven't seen a modern font that GetCharABCWidthsW fails on.
        # For safety, just use the advance as the width.
        if not ttf_font:
            width = advance

            # This hack bumps up the width if the font is italic;
            # this compensates for some common fonts.  It's also a stupid
            # waste of texture memory.
            if self.font.italic:
                width += width // 2
                # Do not enlarge more than the _rect width.
                width = min(width, self._rect.Width)

        # Draw character to bitmap
        gdiplus.GdipGraphicsClear(self._graphics, 0x00000000)
        gdiplus.GdipDrawString(self._graphics,
                               ch,
                               len_ch,
                               self.font._gdipfont,
                               ctypes.byref(rect),
                               fmt,
                               self._brush)

        gdiplus.GdipFlush(self._graphics, 1)
        gdiplus.GdipDeleteStringFormat(fmt)

        bitmap_data = BitmapData()
        gdiplus.GdipBitmapLockBits(
            self._bitmap,
            ctypes.byref(self._rect), ImageLockModeRead, self._format,
            ctypes.byref(bitmap_data))

        # Create buffer for RawImage
        buffer = ctypes.create_string_buffer(
            bitmap_data.Stride * bitmap_data.Height)
        ctypes.memmove(buffer, bitmap_data.Scan0, len(buffer))

        # Unlock data
        gdiplus.GdipBitmapUnlockBits(self._bitmap, ctypes.byref(bitmap_data))

        image = pyglet.image.ImageData(
            width, height,
            "BGRA", buffer, -bitmap_data.Stride)

        glyph = self.font.create_glyph(image)
        # Only pass negative LSB info
        lsb = min(lsb, 0)
        glyph.set_bearings(-self.font.descent, lsb, advance)
        return glyph


class Win32Font(base.Font):
    glyph_renderer_class = GDIPlusGlyphRenderer

    def __init__(
            self,
            name: str, size: float,
            weight: str = "normal", italic: bool = False, stretch: bool = False,
            dpi: int | None = None,
    ) -> None:
        super().__init__()

        bold = weight if weight in (None, True, False) else "bold" in weight

        self.logfont = self.get_logfont(name, size, bold, italic, dpi)
        self.hfont = gdi32.CreateFontIndirectW(ctypes.byref(self.logfont))

        # Create a dummy DC for coordinate mapping
        with device_context(None) as dc:
            metrics = TEXTMETRIC()
            gdi32.SelectObject(dc, self.hfont)
            gdi32.GetTextMetricsA(dc, ctypes.byref(metrics))
            self.ascent = metrics.tmAscent
            self.descent = -metrics.tmDescent
            self.max_glyph_width = metrics.tmMaxCharWidth

    @staticmethod
    def get_logfont(name: str, size: float, bold: bool, italic: bool, dpi: int | None = None) -> LOGFONTW:
        """Get a raw Win32 :py:class:`.LOGFONTW` struct for the given arguments.

        Args:
            name: The name of the font
            size: The font size in points
            bold: Whether to request the font as bold
            italic: Whether to request the font as italic
            dpi: The screen dpi

        Returns:
            LOGFONTW: a ctypes binding of a Win32 LOGFONTW struct
        """

        # Create a dummy DC for coordinate mapping
        with device_context(None) as dc:

            # Default to 96 DPI unless otherwise specified
            if dpi is None:
                dpi = 96
            log_pixels_y = dpi

            # Create LOGFONTW font description struct
            logfont = LOGFONTW()

            # Convert point size to actual device pixels
            logfont.lfHeight = int(-size * log_pixels_y // 72)

            # Configure the LOGFONTW's font properties
            if bold:
                logfont.lfWeight = FW_BOLD
            else:
                logfont.lfWeight = FW_NORMAL
            logfont.lfItalic = italic
            logfont.lfFaceName = name
            logfont.lfQuality = ANTIALIASED_QUALITY

        return logfont


def _get_font_families(font_collection: ctypes.c_void_p) -> Sequence[ctypes.c_void_p]:
    num_count = ctypes.c_int()
    gdiplus.GdipGetFontCollectionFamilyCount(
        font_collection, ctypes.byref(num_count))
    gpfamilies = (ctypes.c_void_p * num_count.value)()
    numFound = ctypes.c_int()
    gdiplus.GdipGetFontCollectionFamilyList(
        font_collection, num_count, gpfamilies, ctypes.byref(numFound))

    return gpfamilies

def _font_exists_in_collection(font_collection: ctypes.c_void_p, name: str) -> bool:
    font_name = ctypes.create_unicode_buffer(32)
    for gpfamily in _get_font_families(font_collection):
        gdiplus.GdipGetFamilyName(gpfamily, font_name, "\0")
        if font_name.value == name:
            return True

    return False


class GDIPlusFont(Win32Font):
    glyph_renderer_class: ClassVar[type[base.GlyphRenderer]] = GDIPlusGlyphRenderer

    _private_collection: ctypes.c_void_p | None = None
    _system_collection: ctypes.c_void_p | None = None

    _default_name = "Arial"

    def __init__(self, name: str, size: float, weight: str = "normal", italic: bool=False, stretch: bool=False,
                 dpi: int | None=None) -> None:
        if not name:
            name = self._default_name

        if stretch:
            warnings.warn("The current font render does not support stretching.")  # noqa: B028

        bold = weight if weight in (None, True, False) else "bold" in weight

        super().__init__(name, size, bold, italic, stretch, dpi)

        self._name = name

        family = ctypes.c_void_p()

        # GDI will add @ in front of a localized font for some Asian languages. However, GDI will also not find it
        # based on that name (???). Here we remove it before checking font collections.
        if name[0] == "@":
            name = name[1:]

        name = ctypes.c_wchar_p(name)

        # Look in private collection first:
        if self._private_collection:
            gdiplus.GdipCreateFontFamilyFromName(name, self._private_collection, ctypes.byref(family))

        # Then in system collection:
        if not family:
            if _debug_font:
                print(f"Warning: Font '{name}' was not found. Defaulting to: {self._default_name}") 

            gdiplus.GdipCreateFontFamilyFromName(name, None, ctypes.byref(family))

        # Nothing found, use default font.
        if not family:
            self._name = self._default_name
            gdiplus.GdipCreateFontFamilyFromName(ctypes.c_wchar_p(self._name), None, ctypes.byref(family))

        if dpi is None:
            unit = UnitPoint
            self.dpi = 96
        else:
            unit = UnitPixel
            size = (size * dpi) // 72
            self.dpi = dpi

        style = 0
        if bold:
            style |= FontStyleBold
        if italic:
            style |= FontStyleItalic
        self._gdipfont = ctypes.c_void_p()
        gdiplus.GdipCreateFont(family, ctypes.c_float(size), style, unit, ctypes.byref(self._gdipfont))
        gdiplus.GdipDeleteFontFamily(family)

    @property
    def name(self) -> str:
        return self._name

    def __del__(self) -> None:
        gdi32.DeleteObject(self.hfont)
        gdiplus.GdipDeleteFont(self._gdipfont)

    @classmethod
    def add_font_data(cls: type[GDIPlusFont], data: bytes) -> None:
        numfonts = ctypes.c_uint32()
        _handle = gdi32.AddFontMemResourceEx(data, len(data), 0, ctypes.byref(numfonts))

        # None means a null handle was returned, ie something went wrong
        if _handle is None:
            raise ctypes.WinError()

        if not cls._private_collection:
            cls._private_collection = ctypes.c_void_p()
            gdiplus.GdipNewPrivateFontCollection(
                ctypes.byref(cls._private_collection))
        gdiplus.GdipPrivateAddMemoryFont(cls._private_collection, data, len(data))

    @classmethod
    def have_font(cls: type[GDIPlusFont], name: str) -> bool:
        # Enumerate the private collection fonts first, as those are most likely to be used.
        if cls._private_collection and _font_exists_in_collection(cls._private_collection, name):
            return True

        # Instead of enumerating all fonts on the system, as there can potentially be thousands, attempt to create
        # the font family with the name. If it does not error (0), then it exists in the system.
        family = ctypes.c_void_p()
        status = gdiplus.GdipCreateFontFamilyFromName(name, None, ctypes.byref(family))
        if status == 0:
            # Delete temp family to prevent leak.
            gdiplus.GdipDeleteFontFamily(family)
            return True

        return False



from collections import deque
import ctypes
import threading
from typing import Deque, Optional, TYPE_CHECKING
import weakref

from pyglet.media.drivers.base import AbstractAudioDriver, AbstractAudioPlayer, MediaEvent
from pyglet.media.drivers.listener import AbstractListener
from pyglet.media.player_worker_thread import PlayerWorkerThread
from pyglet.util import debug_print

from . import lib_pulseaudio as pa
from .interface import PulseAudioMainloop

if TYPE_CHECKING:
    from pyglet.media.codecs import AudioData, AudioFormat, Source
    from pyglet.media.player import Player


_debug = debug_print('debug_media')


class PulseAudioDriver(AbstractAudioDriver):
    def __init__(self) -> None:
        self.mainloop = PulseAudioMainloop()
        self.mainloop.start()
        self.context = None

        self.worker = PlayerWorkerThread()
        self.worker.start()
        self._players = weakref.WeakSet()
        self._listener = PulseAudioListener(self)

    def create_audio_player(self, source: 'Source', player: 'Player') -> 'PulseAudioPlayer':
        assert self.context is not None
        player = PulseAudioPlayer(source, player, self)
        self._players.add(player)
        return player

    def connect(self, server: Optional[bytes] = None) -> None:
        """Connect to pulseaudio server.

        :Parameters:
            `server` : bytes
                Server to connect to, or ``None`` for the default local
                server (which may be spawned as a daemon if no server is
                found).
        """
        # TODO disconnect from old
        assert not self.context, 'Already connected'

        self.context = self.mainloop.create_context()
        self.context.connect(server)

    def dump_debug_info(self):
        print('Client version: ', pa.pa_get_library_version())
        print('Server:         ', self.context.server)
        print('Protocol:       ', self.context.protocol_version)
        print('Server protocol:', self.context.server_protocol_version)
        print('Local context:  ', self.context.is_local and 'Yes' or 'No')

    def delete(self) -> None:
        """Completely shut down pulseaudio client."""
        if self.mainloop is None:
            return

        self.worker.stop()

        with self.mainloop.lock:
            if self.context is not None:
                self.context.delete()
                self.context = None

        self.mainloop.delete()
        self.mainloop = None

    def get_listener(self) -> 'PulseAudioListener':
        return self._listener


class PulseAudioListener(AbstractListener):
    def __init__(self, driver: 'PulseAudioDriver') -> None:
        self.driver = weakref.proxy(driver)

    def _set_volume(self, volume: float) -> None:
        self._volume = volume
        for player in self.driver._players:
            player.set_volume(player._volume)

    def _set_position(self, position):
        self._position = position

    def _set_forward_orientation(self, orientation):
        self._forward_orientation = orientation

    def _set_up_orientation(self, orientation):
        self._up_orientation = orientation


class _AudioDataBuffer:
    def __init__(self, ideal_size: int, comfortable_limit: int) -> None:
        self.available = 0
        self.virtual_write_index = 0
        self._ideal_size = ideal_size
        self._comfortable_limit = comfortable_limit
        self._data: Deque['AudioData'] = deque()
        self._first_read_offset = 0

    def clear(self) -> None:
        self.available = 0
        self.virtual_write_index = 0
        self._data.clear()
        self._first_read_offset = 0

    def get_ideal_refill_size(self, virtual_required: int = 0) -> int:
        virtual_available = self.available - virtual_required
        if virtual_available < self._comfortable_limit:
            return self._ideal_size - virtual_available
        return 0

    def add_data(self, d: 'AudioData') -> None:
        self._data.append(d)
        self.available += d.length
        self.virtual_write_index += d.length

    def memmove(self, target_pointer: int, num_bytes: int) -> int:
        bytes_written = 0
        bytes_remaining = num_bytes
        while bytes_remaining > 0 and self._data:
            cur_audio_data = self._data[0]
            cur_len = cur_audio_data.length - self._first_read_offset
            packet_used = cur_len <= bytes_remaining
            cur_write = min(bytes_remaining, cur_len)
            ctypes.memmove(target_pointer + bytes_written,
                           cur_audio_data.pointer + self._first_read_offset,
                           cur_write)
            bytes_written += cur_write
            bytes_remaining -= cur_write
            if packet_used:
                self._data.popleft()
                self._first_read_offset = 0
            else:
                self._first_read_offset += cur_write

        self.available -= bytes_written

        return bytes_written


class PulseAudioPlayer(AbstractAudioPlayer):
    def __init__(self, source: 'Source', player: 'Player', driver: 'PulseAudioDriver') -> None:
        super().__init__(source, player)
        self.driver = driver

        self._volume = 1.0

        audio_format = source.audio_format
        assert audio_format

        self._latest_timing_info = None
        self._last_clear_read_index = 0

        self._pyglet_source_exhausted = False
        self._pending_bytes = 0

        # PA has a playback buffer it makes requests for via callbacks, but those callbacks
        # can not decode data. The decoding happens in a PlayerWorkerThread as usual, and the
        # result is kept in an _AudioDataBuffer, splitting up the ideal size between
        # them. The data is copied when the PulseAudio buffer runs low, at which point the
        # thread will ensure refilling of the _AudioDataBuffer.
        ideal_size = audio_format.align_ceil(self._buffered_data_ideal_size // 2)
        comf_limit = audio_format.align_ceil(self._buffered_data_comfortable_limit // 2)

        self._audio_data_buffer = _AudioDataBuffer(ideal_size, comf_limit)

        # A lock that should be held whenever the audio data buffer is accessed, or
        # any variables really, if they are shared between PA callbacks and the rest of
        # implemented methods.
        # Should prevent The PA callback from interfering with the work method.
        # Don't ever acquire the PA mainloop lock when this is held, might risk a deadlock
        # if a callback runs at an unfortunate time.
        self._audio_data_lock = threading.Lock()

        self._has_underrun = False

        with driver.mainloop.lock:
            self.stream = driver.context.create_stream(audio_format)
            self.stream.set_write_callback(self._write_callback)
            self.stream.set_underflow_callback(self._underflow_callback)
            self.stream.connect_playback(ideal_size, comf_limit)
            assert self.stream.is_ready

        assert _debug('PulseAudioPlayer: __init__ finished')

    def _write_callback(self, _stream, nbytes: int, _userdata) -> None:
        # Called from within PA thread
        assert _debug(f'PulseAudioPlayer: Write requested, {nbytes}B')
        assert self.source.audio_format.align(nbytes) == nbytes

        with self._audio_data_lock:
            if self._audio_data_buffer.available > 0:
                written = self._write_to_stream(nbytes)
                if (unfulfilled := nbytes - written) > 0:
                    self._pending_bytes = unfulfilled
            else:
                self._pending_bytes = nbytes

        self.stream.mainloop.signal()

    def _underflow_callback(self, _stream, _userdata) -> None:
        # Called from within PA thread
        assert _debug('PulseAudioPlayer: underflow')
        with self._audio_data_lock:
            if self._pyglet_source_exhausted and self._audio_data_buffer.available == 0:
                MediaEvent('on_eos').sync_dispatch_to_player(self.player)
            self._has_underrun = True
        self.stream.mainloop.signal()

    def _maybe_fill_audio_data_buffer(self) -> None:
        # PA as opposed to the other backends works on requests which are relatively small (or on a
        # polling model not used here which also requires the client to adjust to an ideal remaining
        # space), so this chops up AudioData in an attempt of not hitting source.get_audio_data too
        # often.
        # Hold the audio_data_lock when calling this.
        if self._pyglet_source_exhausted:
            return

        refill_size = self._audio_data_buffer.get_ideal_refill_size(self._pending_bytes)
        if refill_size == 0:
            return

        self._audio_data_lock.release()

        refill_size = self.source.audio_format.align(refill_size)
        assert _debug(f'PulseAudioPlayer: Getting {refill_size}B of audio data')
        new_data = self._get_and_compensate_audio_data(refill_size, self._get_read_index())

        self._audio_data_lock.acquire()

        if new_data is None:
            self._pyglet_source_exhausted = True
            if self._has_underrun:
                MediaEvent('on_eos').sync_dispatch_to_player(self.player)
        else:
            self._audio_data_buffer.add_data(new_data)
            self.append_events(self._audio_data_buffer.virtual_write_index, new_data.events)

    def _write_to_stream(self, nbytes: int) -> int:
        data_ptr, bytes_accepted = self.stream.begin_write(nbytes)

        bytes_written = self._audio_data_buffer.memmove(data_ptr.value, bytes_accepted)

        if bytes_written == 0:
            self.stream.cancel_write()
        else: # elif bytes_written <= bytes_accepted:
            self.stream.write(data_ptr, bytes_written, pa.PA_SEEK_RELATIVE)

        assert _debug(f'PulseAudioPlayer: Wrote {bytes_written}/{nbytes}')
        return bytes_written

    def _update_and_get_timing_info(self) -> Optional[pa.pa_timing_info]:
        self.stream.update_timing_info().wait().delete()
        return self.stream.get_timing_info()

    def _maybe_write_pending(self) -> None:
        with self._audio_data_lock:
            if self._pending_bytes == 0 or self._audio_data_buffer.available == 0:
                return

            written = self._write_to_stream(self._pending_bytes)
            self._pending_bytes -= written
            if not self._has_underrun:
                return

            self._has_underrun = False

        # If the stream has underrun, trigger it again for immediate playback.
        # Unsure whether this is really needed, but better be safe.
        # Make sure to not hold the audio data lock, as `wait` reenters the PA loop.
        self.stream.trigger().wait().delete()

    def work(self) -> None:
        with self.driver.mainloop.lock:
            self._maybe_write_pending()
            self._latest_timing_info = self._update_and_get_timing_info()

        self.dispatch_media_events(self._get_read_index())
        with self._audio_data_lock:
            self._maybe_fill_audio_data_buffer()
        with self.driver.mainloop.lock:
            self._maybe_write_pending()

    def delete(self) -> None:
        assert _debug('PulseAudioPlayer.delete')
        self.driver.worker.remove(self)

        if self.driver.mainloop is None:
            assert _debug('PulseAudioPlayer.delete: PulseAudioDriver already deleted.')
            # This is fine otherwise. If the mainloop's gone, the context is gone too,
            # having cleaned up all its streams.
        else:
            with self.driver.mainloop.lock:
                self.stream.delete()
                self.stream = None

    def clear(self) -> None:
        assert _debug('PulseAudioPlayer.clear')
        super().clear()

        # Do not reset pending_bytes as PA doesn't always seem to issue new write
        # requests despite being flushed. Should still end up fine since PA replies with
        # how many bytes it actually accepts in `_write_to_stream`.
        # self._pending_bytes = 0
        self._pyglet_source_exhausted = False
        self._audio_data_buffer.clear()
        self._has_underrun = False

        with self.stream.mainloop.lock:
            # Just hope that the read index is frozen while we're paused.
            ti = self._update_and_get_timing_info()
            assert not ti.read_index_corrupt
            self._last_clear_read_index = ti.read_index
            self.stream.flush().wait().delete()
            self.stream.prebuf().wait().delete()

    def play(self) -> None:
        assert _debug('PulseAudioPlayer.play')

        with self.stream.mainloop.lock:
            if self.stream.is_corked():
                self.stream.resume().wait().delete()
            assert not self.stream.is_corked()

        self.driver.worker.add(self)

    def stop(self) -> None:
        assert _debug('PulseAudioPlayer.stop')
        self.driver.worker.remove(self)

        with self.stream.mainloop.lock:
            self.stream.pause().wait().delete()

    def get_play_cursor(self) -> int:
        return self._get_read_index()

    def _get_read_index(self) -> int:
        if (t_info := self._latest_timing_info) is None:
            return 0

        read_idx = t_info.read_index - self._last_clear_read_index
        # bps = self.source.audio_format.bytes_per_second
        # (t_info.transport_usec / 1000000.0) * bps -
        # (t_info.sink_usec / 1000000.0) * bps

        assert _debug(f'_get_read_index -> {read_idx}')
        return read_idx

    def set_volume(self, volume: float) -> None:
        self._volume = volume

        if self.stream:
            driver = self.driver
            volume *= driver._listener._volume
            with driver.context.mainloop.lock:
                driver.context.set_input_volume(self.stream, volume).wait().delete()

    def set_pitch(self, pitch):
        with self.stream.mainloop.lock:
            sample_rate = self.stream.get_sample_spec().rate
            self.stream.update_sample_rate(int(pitch * sample_rate)).wait().delete()

    def prefill_audio(self):
        self.work()



import ctypes
import pyglet
from pyglet.input.base import Device, DeviceOpenException
from pyglet.input.base import Button, RelativeAxis, AbsoluteAxis
from pyglet.libs.x11 import xlib
from pyglet.util import asstr

try:
    from pyglet.libs.x11 import xinput as xi

    _have_xinput = True
except:
    _have_xinput = False


def ptr_add(ptr, offset):
    address = ctypes.addressof(ptr.contents) + offset
    return ctypes.pointer(type(ptr.contents).from_address(address))


class DeviceResponder:
    def _key_press(self, e):
        pass

    def _key_release(self, e):
        pass

    def _button_press(self, e):
        pass

    def _button_release(self, e):
        pass

    def _motion(self, e):
        pass

    def _proximity_in(self, e):
        pass

    def _proximity_out(self, e):
        pass


class XInputDevice(DeviceResponder, Device):
    def __init__(self, display, device_info):
        super().__init__(display, asstr(device_info.name))

        self._device_id = device_info.id
        self._device = None

        # Read device info
        self.buttons = []
        self.keys = []
        self.axes = []

        ptr = device_info.inputclassinfo
        for i in range(device_info.num_classes):
            cp = ctypes.cast(ptr, ctypes.POINTER(xi.XAnyClassInfo))
            cls_class = getattr(cp.contents, 'class')

            if cls_class == xi.KeyClass:
                cp = ctypes.cast(ptr, ctypes.POINTER(xi.XKeyInfo))
                self.min_keycode = cp.contents.min_keycode
                num_keys = cp.contents.num_keys
                for i in range(num_keys):
                    self.keys.append(Button('key%d' % i))

            elif cls_class == xi.ButtonClass:
                cp = ctypes.cast(ptr, ctypes.POINTER(xi.XButtonInfo))
                num_buttons = cp.contents.num_buttons
                # Pointer buttons start at index 1, with 0 as 'AnyButton'
                for i in range(num_buttons + 1):
                    self.buttons.append(Button('button%d' % i))

            elif cls_class == xi.ValuatorClass:
                cp = ctypes.cast(ptr, ctypes.POINTER(xi.XValuatorInfo))
                num_axes = cp.contents.num_axes
                mode = cp.contents.mode
                axes = ctypes.cast(cp.contents.axes,
                                   ctypes.POINTER(xi.XAxisInfo))
                for i in range(num_axes):
                    axis = axes[i]
                    if mode == xi.Absolute:
                        self.axes.append(AbsoluteAxis('axis%d' % i, axis.min_value, axis.max_value))
                    elif mode == xi.Relative:
                        self.axes.append(RelativeAxis('axis%d' % i))

            cls = cp.contents
            ptr = ptr_add(ptr, cls.length)

        self.controls = self.buttons + self.keys + self.axes

        # Can't detect proximity class event without opening device.  Just
        # assume there is the possibility of a control if there are any axes.
        if self.axes:
            self.proximity_control = Button('proximity')
            self.controls.append(self.proximity_control)
        else:
            self.proximity_control = None

    def get_controls(self):
        return self.controls

    def open(self, window=None, exclusive=False):
        # Checks for is_open and raises if already open.
        # TODO allow opening on multiple windows.
        super().open(window, exclusive)

        if window is None:
            self._is_open = False
            raise DeviceOpenException('XInput devices require a window')

        if window.display._display != self.display._display:
            self._is_open = False
            raise DeviceOpenException('Window and device displays differ')

        if exclusive:
            self._is_open = False
            raise DeviceOpenException('Cannot open XInput device exclusive')

        self._device = xi.XOpenDevice(self.display._display, self._device_id)
        if not self._device:
            self._is_open = False
            raise DeviceOpenException('Cannot open device')

        self._install_events(window)

    def close(self):
        super().close()

        if not self._device:
            return

        # TODO: uninstall events
        xi.XCloseDevice(self.display._display, self._device)

    def _install_events(self, window):
        dispatcher = XInputWindowEventDispatcher.get_dispatcher(window)
        dispatcher.open_device(self._device_id, self._device, self)

    # DeviceResponder interface

    def _key_press(self, e):
        self.keys[e.keycode - self.min_keycode].value = True

    def _key_release(self, e):
        self.keys[e.keycode - self.min_keycode].value = False

    def _button_press(self, e):
        self.buttons[e.button].value = True

    def _button_release(self, e):
        self.buttons[e.button].value = False

    def _motion(self, e):
        for i in range(e.axes_count):
            self.axes[i].value = e.axis_data[i]

    def _proximity_in(self, e):
        if self.proximity_control:
            self.proximity_control.value = True

    def _proximity_out(self, e):
        if self.proximity_control:
            self.proximity_control.value = False


class XInputWindowEventDispatcher:
    def __init__(self, window):
        self.window = window
        self._responders = {}

    @staticmethod
    def get_dispatcher(window):
        try:
            dispatcher = window.__xinput_window_event_dispatcher
        except AttributeError:
            dispatcher = window.__xinput_window_event_dispatcher = XInputWindowEventDispatcher(window)
        return dispatcher

    def set_responder(self, device_id, responder):
        self._responders[device_id] = responder

    def remove_responder(self, device_id):
        del self._responders[device_id]

    def open_device(self, device_id, device, responder):
        self.set_responder(device_id, responder)
        device = device.contents
        if not device.num_classes:
            return

        # Bind matching extended window events to bound instance methods
        # on this object.
        #
        # This is inspired by test.c of xinput package by Frederic
        # Lepied available at x.org.
        #
        # In C, this stuff is normally handled by the macro DeviceKeyPress and
        # friends. Since we don't have access to those macros here, we do it
        # this way.
        events = []

        def add(class_info, event, handler):
            _type = class_info.event_type_base + event
            _class = device_id << 8 | _type
            events.append(_class)
            self.window._event_handlers[_type] = handler

        for i in range(device.num_classes):
            class_info = device.classes[i]
            if class_info.input_class == xi.KeyClass:
                add(class_info, xi._deviceKeyPress, self._event_xinput_key_press)
                add(class_info, xi._deviceKeyRelease, self._event_xinput_key_release)

            elif class_info.input_class == xi.ButtonClass:
                add(class_info, xi._deviceButtonPress, self._event_xinput_button_press)
                add(class_info, xi._deviceButtonRelease, self._event_xinput_button_release)

            elif class_info.input_class == xi.ValuatorClass:
                add(class_info, xi._deviceMotionNotify, self._event_xinput_motion)

            elif class_info.input_class == xi.ProximityClass:
                add(class_info, xi._proximityIn, self._event_xinput_proximity_in)
                add(class_info, xi._proximityOut, self._event_xinput_proximity_out)

            elif class_info.input_class == xi.FeedbackClass:
                pass

            elif class_info.input_class == xi.FocusClass:
                pass

            elif class_info.input_class == xi.OtherClass:
                pass

        array = (xi.XEventClass * len(events))(*events)
        xi.XSelectExtensionEvent(self.window._x_display, self.window._window, array, len(array))

    @pyglet.window.xlib.XlibEventHandler(0)
    def _event_xinput_key_press(self, ev):
        e = ctypes.cast(ctypes.byref(ev), ctypes.POINTER(xi.XDeviceKeyEvent)).contents

        device = self._responders.get(e.deviceid)
        if device is not None:
            device._key_press(e)

    @pyglet.window.xlib.XlibEventHandler(0)
    def _event_xinput_key_release(self, ev):
        e = ctypes.cast(ctypes.byref(ev), ctypes.POINTER(xi.XDeviceKeyEvent)).contents

        device = self._responders.get(e.deviceid)
        if device is not None:
            device._key_release(e)

    @pyglet.window.xlib.XlibEventHandler(0)
    def _event_xinput_button_press(self, ev):
        e = ctypes.cast(ctypes.byref(ev), ctypes.POINTER(xi.XDeviceButtonEvent)).contents

        device = self._responders.get(e.deviceid)
        if device is not None:
            device._button_press(e)

    @pyglet.window.xlib.XlibEventHandler(0)
    def _event_xinput_button_release(self, ev):
        e = ctypes.cast(ctypes.byref(ev), ctypes.POINTER(xi.XDeviceButtonEvent)).contents

        device = self._responders.get(e.deviceid)
        if device is not None:
            device._button_release(e)

    @pyglet.window.xlib.XlibEventHandler(0)
    def _event_xinput_motion(self, ev):
        e = ctypes.cast(ctypes.byref(ev), ctypes.POINTER(xi.XDeviceMotionEvent)).contents

        device = self._responders.get(e.deviceid)
        if device is not None:
            device._motion(e)

    @pyglet.window.xlib.XlibEventHandler(0)
    def _event_xinput_proximity_in(self, ev):
        e = ctypes.cast(ctypes.byref(ev), ctypes.POINTER(xi.XProximityNotifyEvent)).contents

        device = self._responders.get(e.deviceid)
        if device is not None:
            device._proximity_in(e)

    @pyglet.window.xlib.XlibEventHandler(-1)
    def _event_xinput_proximity_out(self, ev):
        e = ctypes.cast(ctypes.byref(ev), ctypes.POINTER(xi.XProximityNotifyEvent)).contents

        device = self._responders.get(e.deviceid)
        if device is not None:
            device._proximity_out(e)


def _check_extension(display):
    major_opcode = ctypes.c_int()
    first_event = ctypes.c_int()
    first_error = ctypes.c_int()
    xlib.XQueryExtension(display._display,
                         b'XInputExtension',
                         ctypes.byref(major_opcode),
                         ctypes.byref(first_event),
                         ctypes.byref(first_error))
    return bool(major_opcode.value)


def get_devices(display=None):
    if display is None:
        display = pyglet.display.get_display()

    if not _have_xinput or not _check_extension(display):
        return []

    devices = []
    count = ctypes.c_int(0)
    device_list = xi.XListInputDevices(display._display, count)

    for i in range(count.value):
        device_info = device_list[i]
        devices.append(XInputDevice(display, device_info))

    xi.XFreeDeviceList(device_list)

    return devices



"""Wrapper around the Linux FontConfig library. Used to find available fonts."""
from __future__ import annotations

from collections import OrderedDict
from ctypes import CDLL, Structure, Union, byref, c_char_p, c_double, c_int, c_uint, c_void_p, POINTER
from typing import TYPE_CHECKING

from pyglet.font.base import FontException
from pyglet.lib import load_library
from pyglet.util import asbytes, asstr

if TYPE_CHECKING:
    from pyglet.font.freetype_lib import FT_Face

# fontconfig library definitions

(FcResultMatch,
 FcResultNoMatch,
 FcResultTypeMismatch,
 FcResultNoId,
 FcResultOutOfMemory) = range(5)

FcResult = c_int

FC_FAMILY = asbytes("family")
FC_SIZE = asbytes("size")
FC_SLANT = asbytes("slant")
FC_WEIGHT = asbytes("weight")
FC_FT_FACE = asbytes("ftface")
FC_FILE = asbytes("file")
FC_WIDTH = asbytes("width")

FC_WEIGHT_THIN = 10
FC_WEIGHT_EXTRALIGHT = 40
FC_WEIGHT_ULTRALIGHT = FC_WEIGHT_EXTRALIGHT
FC_WEIGHT_LIGHT = 50
FC_WEIGHT_DEMILIGHT = 55
FC_WEIGHT_SEMILIGHT = FC_WEIGHT_DEMILIGHT
FC_WEIGHT_BOOK = 75
FC_WEIGHT_REGULAR = 80
FC_WEIGHT_NORMAL = FC_WEIGHT_REGULAR
FC_WEIGHT_MEDIUM = 100
FC_WEIGHT_DEMIBOLD = 180
FC_WEIGHT_SEMIBOLD = FC_WEIGHT_DEMIBOLD
FC_WEIGHT_BOLD = 200
FC_WEIGHT_EXTRABOLD = 205
FC_WEIGHT_ULTRABOLD = FC_WEIGHT_EXTRABOLD
FC_WEIGHT_BLACK = 210
FC_WEIGHT_HEAVY = FC_WEIGHT_BLACK
FC_WEIGHT_EXTRABLACK = 215
FC_WEIGHT_ULTRABLACK = FC_WEIGHT_EXTRABLACK

name_to_weight = {
    True: FC_WEIGHT_BOLD,  # Temporary alias for attributed text
    False: FC_WEIGHT_NORMAL,  # Temporary alias for attributed text
    None: FC_WEIGHT_NORMAL,  # Temporary alias for attributed text
    "thin": FC_WEIGHT_THIN,
    "extralight": FC_WEIGHT_EXTRALIGHT,
    "ultralight": FC_WEIGHT_ULTRALIGHT,
    "light": FC_WEIGHT_LIGHT,
    "semilight": FC_WEIGHT_SEMILIGHT,
    "normal": FC_WEIGHT_NORMAL,
    "regular": FC_WEIGHT_REGULAR,
    "medium": FC_WEIGHT_MEDIUM,
    "demibold": FC_WEIGHT_DEMIBOLD,
    "semibold": FC_WEIGHT_SEMIBOLD,
    "bold": FC_WEIGHT_BOLD,
    "extrabold": FC_WEIGHT_EXTRABOLD,
    "ultrabold": FC_WEIGHT_ULTRABOLD,
    "black": FC_WEIGHT_BLACK,
    "heavy": FC_WEIGHT_HEAVY,
    "extrablack": FC_WEIGHT_EXTRABLACK,
}

weight_to_name = {
    None: 'normal',
    0: 'thin',
    40: 'extralight',
    50: 'light',
    55: 'semilight',
    80: 'normal',
    100: 'medium',
    180: 'semibold',
    200: 'bold',
    205: 'extrabold',
    210: 'black',
    215: 'ultrabold',
}


FC_WIDTH_ULTRACONDENSED = 50
FC_WIDTH_EXTRACONDENSED = 63
FC_WIDTH_CONDENSED = 75
FC_WIDTH_SEMICONDENSED = 87
FC_WIDTH_NORMAL = 100
FC_WIDTH_SEMIEXPANDED = 113
FC_WIDTH_EXPANDED = 125
FC_WIDTH_EXTRAEXPANDED = 150
FC_WIDTH_ULTRAEXPANDED = 200


name_to_stretch = {
    None: FC_WIDTH_NORMAL,
    False: FC_WIDTH_NORMAL,
    "undefined": FC_WIDTH_NORMAL,
    "ultracondensed": FC_WIDTH_ULTRACONDENSED,
    "extracondensed": FC_WIDTH_EXTRACONDENSED,
    "condensed": FC_WIDTH_CONDENSED,
    "semicondensed": FC_WIDTH_SEMICONDENSED,
    "normal": FC_WIDTH_NORMAL,
    "medium": FC_WIDTH_NORMAL,
    "semiexpanded": FC_WIDTH_SEMIEXPANDED,
    "expanded": FC_WIDTH_EXPANDED,
    "extraexpanded": FC_WIDTH_EXTRAEXPANDED,
    "ultraexpanded": FC_WIDTH_ULTRAEXPANDED,
}

stretch_to_name = {
    None: "normal",
    False: "normal",
    FC_WIDTH_ULTRACONDENSED : "ultracondensed",
    FC_WIDTH_EXTRACONDENSED: "extracondensed",
    FC_WIDTH_CONDENSED: "condensed",
    FC_WIDTH_SEMICONDENSED: "semicondensed",
    FC_WIDTH_NORMAL: "normal",
    FC_WIDTH_SEMIEXPANDED: "semiexpanded",
    FC_WIDTH_EXPANDED:  "expanded",
    FC_WIDTH_EXTRAEXPANDED: "extraexpanded",
    FC_WIDTH_ULTRAEXPANDED: "narrow",
}

FC_SLANT_ROMAN = 0
FC_SLANT_ITALIC = 100
FC_SLANT_OBLIQUE = 110

(FcTypeVoid,
 FcTypeInteger,
 FcTypeDouble,
 FcTypeString,
 FcTypeBool,
 FcTypeMatrix,
 FcTypeCharSet,
 FcTypeFTFace,
 FcTypeLangSet) = range(9)
FcType = c_int

(FcMatchPattern,
 FcMatchFont) = range(2)
FcMatchKind = c_int


class _FcValueUnion(Union):
    _fields_ = [
        ("s", c_char_p),
        ("i", c_int),
        ("b", c_int),
        ("d", c_double),
        ("m", c_void_p),
        ("c", c_void_p),
        ("f", c_void_p),
        ("p", c_void_p),
        ("l", c_void_p),
    ]


class FcValue(Structure):
    _fields_ = [
        ("type", FcType),
        ("u", _FcValueUnion),
    ]


# End of library definitions


class FontConfig:
    _search_cache: OrderedDict[tuple[str, float, str, bool, str], FontConfigSearchResult]
    _fontconfig: CDLL | None

    def __init__(self) -> None:
        self._fontconfig = self._load_fontconfig_library()
        assert self._fontconfig is not None
        self._search_cache = OrderedDict()
        self._cache_size = 20

    def dispose(self) -> None:
        while len(self._search_cache) > 0:
            k, v = self._search_cache.popitem()
            v.dispose()

        self._fontconfig.FcFini()
        self._fontconfig = None

    def create_search_pattern(self) -> FontConfigSearchPattern:
        return FontConfigSearchPattern(self._fontconfig)

    def style_from_face(self, font_face: FT_Face):
        pattern = self._fontconfig.FcFreeTypeQueryFace(font_face, None, 0, None)
        result = FontConfigSearchResult(self._fontconfig, pattern)
        return result.weight, result.italic, result.stretch

    def find_font(self, name: str, size: float = 12, weight: str = "normal",
                  italic: bool = False, stretch: str = "normal") -> FontConfigSearchResult:
        if result := self._get_from_search_cache(name, size, weight, italic, stretch):
            return result

        search_pattern = self.create_search_pattern()
        search_pattern.name = name
        search_pattern.size = size
        search_pattern.weight = weight
        search_pattern.italic = italic
        search_pattern.stretch = stretch

        result = search_pattern.match()
        self._add_to_search_cache(search_pattern, result)
        search_pattern.dispose()
        return result

    def have_font(self, name: str) -> bool:
        if result := self.find_font(name):
            # Check the name matches, fontconfig can return a default
            if name and result.name and result.name.lower() != name.lower():
                return False
            return True

        return False

    def char_index(self, ft_face: FT_Face, character: str) -> int:
        return self._fontconfig.FcFreeTypeCharIndex(ft_face, ord(character))

    def _add_to_search_cache(self, search_pattern: FontConfigSearchPattern,
                             result_pattern: FontConfigSearchResult) -> None:
        self._search_cache[(search_pattern.name,
                            search_pattern.size,
                            search_pattern.weight,
                            search_pattern.italic,
                            search_pattern.stretch)] = result_pattern
        if len(self._search_cache) > self._cache_size:
            self._search_cache.popitem(last=False)[1].dispose()

    def _get_from_search_cache(self, name: str, size: float, weight: str,
                               italic: bool, stretch: str) -> FontConfigSearchResult | None:
        result = self._search_cache.get((name, size, weight, italic, stretch), None)

        if result and result.is_valid:
            return result

        return None

    @staticmethod
    def _load_fontconfig_library() -> CDLL:
        fontconfig = load_library("fontconfig")
        fontconfig.FcInit()

        fontconfig.FcPatternBuild.restype = c_void_p
        fontconfig.FcPatternCreate.restype = c_void_p
        fontconfig.FcFontMatch.restype = c_void_p
        fontconfig.FcFreeTypeCharIndex.restype = c_uint

        fontconfig.FcPatternAddDouble.argtypes = [c_void_p, c_char_p, c_double]
        fontconfig.FcPatternAddInteger.argtypes = [c_void_p, c_char_p, c_int]
        fontconfig.FcPatternAddString.argtypes = [c_void_p, c_char_p, c_char_p]
        fontconfig.FcConfigSubstitute.argtypes = [c_void_p, c_void_p, c_int]
        fontconfig.FcDefaultSubstitute.argtypes = [c_void_p]
        fontconfig.FcFontMatch.argtypes = [c_void_p, c_void_p, c_void_p]
        fontconfig.FcPatternDestroy.argtypes = [c_void_p]

        fontconfig.FcPatternGetFTFace.argtypes = [c_void_p, c_char_p, c_int, c_void_p]
        fontconfig.FcPatternGet.argtypes = [c_void_p, c_char_p, c_int, c_void_p]

        fontconfig.FcFreeTypeQueryFace.argtypes = [c_void_p, c_char_p, c_int, POINTER(c_int)]
        fontconfig.FcFreeTypeQueryFace.restype = c_void_p

        return fontconfig


class FontConfigPattern:
    def __init__(self, fontconfig: CDLL, pattern: c_void_p | None = None) -> None:
        self._fontconfig = fontconfig
        self._pattern = pattern

    @property
    def is_valid(self) -> bool:
        return bool(self._fontconfig and self._pattern)

    def _create(self) -> None:
        assert not self._pattern
        assert self._fontconfig
        self._pattern = self._fontconfig.FcPatternCreate()

    def _destroy(self) -> None:
        assert self._pattern
        assert self._fontconfig
        self._fontconfig.FcPatternDestroy(self._pattern)
        self._pattern = None

    @staticmethod
    def _italic_to_slant(italic: bool) -> int:
        return FC_SLANT_ITALIC if italic else FC_SLANT_ROMAN

    def _set_string(self, name: bytes, value: str) -> None:
        assert self._pattern
        assert name
        assert self._fontconfig

        if not value:
            return

        value = value.encode("utf8")

        self._fontconfig.FcPatternAddString(self._pattern, name, asbytes(value))

    def _set_double(self, name: bytes, value: float) -> None:
        assert self._pattern
        assert name
        assert self._fontconfig

        if not value:
            return

        self._fontconfig.FcPatternAddDouble(self._pattern, name, c_double(value))

    def _set_integer(self, name: bytes, value: int) -> None:
        assert self._pattern
        assert name
        assert self._fontconfig

        if not value:
            return

        self._fontconfig.FcPatternAddInteger(self._pattern, name, c_int(value))

    def _get_value(self, name: bytes) -> FcValue | None:
        assert self._pattern
        assert name
        assert self._fontconfig

        value = FcValue()
        result: FcResult = self._fontconfig.FcPatternGet(self._pattern, name, 0, byref(value))
        if _handle_fcresult(result):
            return value

        return None

    def _get_string(self, name: bytes) -> str | None:
        value = self._get_value(name)

        if value and value.type == FcTypeString:
            return asstr(value.u.s)

        return None

    def _get_face(self, name: bytes) -> FT_Face | None:
        value = self._get_value(name)

        if value and value.type == FcTypeFTFace:
            return value.u.f

        return None

    def _get_integer(self, name: bytes) -> int | None:
        value = self._get_value(name)

        if value and value.type == FcTypeInteger:
            return value.u.i

        return None

    def _get_double(self, name: bytes) -> int | None:
        value = self._get_value(name)

        if value and value.type == FcTypeDouble:
            return value.u.d

        return None


class FontConfigSearchPattern(FontConfigPattern):
    size: float | None
    italic: bool
    weight: str
    name: str | None
    stretch: str

    def __init__(self, fontconfig: CDLL, pattern: c_void_p | None = None) -> None:
        super().__init__(fontconfig, pattern)

        self.name = None
        self.weight = "normal"
        self.italic = False
        self.size = None
        self.stretch = 'normal'

    def match(self) -> FontConfigSearchResult | None:
        self._prepare_search_pattern()
        result_pattern = self._get_match()

        if result_pattern:
            return FontConfigSearchResult(self._fontconfig, result_pattern)

        return None

    def _prepare_search_pattern(self) -> None:
        if self._pattern is None:
            self._create()
        self._set_string(FC_FAMILY, self.name)
        self._set_double(FC_SIZE, self.size)
        self._set_double(FC_WEIGHT, name_to_weight[self.weight])
        self._set_integer(FC_SLANT, self._italic_to_slant(self.italic))
        self._set_integer(FC_WIDTH, name_to_stretch[self.stretch])

        self._substitute_defaults()

    def _substitute_defaults(self) -> None:
        assert self._pattern
        assert self._fontconfig

        self._fontconfig.FcConfigSubstitute(None, self._pattern, FcMatchPattern)
        self._fontconfig.FcDefaultSubstitute(self._pattern)

    def _get_match(self) -> c_void_p | None:
        assert self._pattern
        assert self._fontconfig

        match_result = FcResult()
        match_pattern = self._fontconfig.FcFontMatch(0, self._pattern, byref(match_result))

        if _handle_fcresult(match_result.value):
            return match_pattern

        return None

    def dispose(self) -> None:
        self._destroy()


class FontConfigSearchResult(FontConfigPattern):
    def __init__(self, fontconfig: CDLL, result_pattern: c_void_p | None) -> None:
        super().__init__(fontconfig, result_pattern)

    @property
    def name(self) -> str:
        return self._get_string(FC_FAMILY)

    @property
    def size(self) -> int:
        return self._get_double(FC_SIZE)

    @property
    def weight(self) -> str:
        return weight_to_name[self._get_double(FC_WEIGHT)]

    @property
    def italic(self) -> bool:
        return self._get_integer(FC_SLANT) == FC_SLANT_ITALIC

    @property
    def face(self) -> FT_Face:
        return self._get_face(FC_FT_FACE)

    @property
    def stretch(self) -> str:
        return stretch_to_name[self._get_integer(FC_WIDTH)]

    @property
    def file(self) -> str:
        return self._get_string(FC_FILE)

    def dispose(self) -> None:
        self._destroy()


def _handle_fcresult(result: int) -> bool | None:
    if result == FcResultMatch:
        return True
    if result in (FcResultNoMatch, FcResultTypeMismatch, FcResultNoId):
        return False
    if result == FcResultOutOfMemory:
        msg = "FontConfig ran out of memory."
        raise FontException(msg)
    return None


_fontconfig_instance = None


def get_fontconfig() -> FontConfig:
    global _fontconfig_instance  # noqa: PLW0603
    if not _fontconfig_instance:
        _fontconfig_instance = FontConfig()
    return _fontconfig_instance
